/*
 * (INCA) ASC UART support
 */

#include <config.h>

#define MON_OFFSET   0xbf020000
#define MON_HAPPY      (MON_OFFSET + 0x0000)
#define MON_SAD        (MON_OFFSET + 0x0008)
#define MON_SRAM_SIZE  (MON_OFFSET + 0x0010)
#define MON_INTERRUPT  (MON_OFFSET + 0x0018)
#define MON_PUTCH      (MON_OFFSET + 0x0020)
#define MON_VCDTRACE   (MON_OFFSET + 0x0028)
#define MON_TIME       (MON_OFFSET + 0x0030)

#include <common.h>
#include <asm/addrspace.h>
#include <asm/inca-ip2.h>
#include "asc.h"

#define ASC_FIFO_PRESENT

#define SET_BIT(reg, mask)                  reg |= (mask)
#define CLEAR_BIT(reg, mask)                reg &= (~mask)
#define CLEAR_BITS(reg, mask)               CLEAR_BIT(reg, mask)
#define SET_BITS(reg, mask)                 SET_BIT(reg, mask)
#define SET_BITFIELD(reg, mask, off, val)   {reg &= (~mask); reg |= (val << off);}

/* pointer to ASC register base address */
static volatile inca2Asc_t *pAsc = (inca2Asc_t *)INCA_IP2_ASC0;

extern uint incaip2_get_fbs1clk(void);

static int serial_setopt (void);
void serial_setbrg (void);


/******************************************************************************
*
* serial_init - initialize an INCA-IP2 ASC channel
*
* This routine initializes the number of data bits, parity
* and set the selected baud rate. Interrupts are disabled.
* Set the modem control signals if the option is selected.
*
* RETURNS: N/A
*/

int serial_init (void)
{
#ifdef EMULATION
	u32 reg;

	/* we have to set PMS.GEN13 bit to enable an ASC0 device*/
	*INCA_IP2_PMS_PMS_GEN |= INCA_IP2_PMS_PMS_GEN_ASC0 | INCA_IP2_PMS_PMS_GEN_ASC1;

	/* Enable ASC1 */
	*INCA_IP2_GPIO_ALTSEL1 = 0x0000003c;
	*INCA_IP2_GPIO_ALTSEL2 = 0x04000000;

	/* Set DIR for ASC0 and ASC1 (TX=Out) */
	*INCA_IP2_GPIO_DIR |= 0x00820000;

	/* and we have to set CLC register*/
	pAsc->asc_clc = 0x00000100;

	/* initialy we are in async mode */
	pAsc->asc_mcon = ASCMCON_M_8ASYNC | 0x00c00000 ;

	/* select input port */
	pAsc->asc_pisel = (CONSOLE_TTY & 0x1);

	/* TXFIFO's filling level */
	/* enable TXFIFO */
	pAsc->asc_txfcon |= ASCTXFCON_TXFEN | ASCTXFCON_TXFIFO1;

	/* RXFIFO's filling level */
	/* enable RXFIFO */
	//	pAsc->asc_rxfcon |= ASCRXFCON_RXFEN | ASCRXFCON_RXFIFO1;
	pAsc->asc_rxfcon |= ASCRXFCON_RXFIFO1;

	/* enable error signals */
	reg = pAsc->asc_mcon;
	reg = reg | ASCMCON_FEN;
	pAsc->asc_mcon = reg;

	/* Clear all error interrupts and disable receiver */
	pAsc->asc_whbstate=0x000000fd;

	/* acknowledge ASC interrupts */
	ASC_INTERRUPTS_CLEAR(INCA2ASC_IRQ_LINE_ALL);

	/* disable ASC interrupts */
	ASC_INTERRUPTS_DISABLE(INCA2ASC_IRQ_LINE_ALL);

	/* set baud rate */
	serial_setbrg();

	if ((u32) pAsc == (u32) INCA_IP2_ASC0) {
		/* Set FIFO to single byte forward mode */
		pAsc->asc_eomcon = 0x00010300;
	}
   
	/* Enable receiver */
	pAsc->asc_whbstate=0x00000002;

	return 0;
#else
	u32 reg;

	/* we have to set PMS.GEN13 bit to enable an ASC0 device*/
	*INCA_IP2_PMS_PMS_GEN |= INCA_IP2_PMS_PMS_GEN_ASC0;
#if 0 /* Only for ASC1 */ 
	/* Enable ASC1 */
	*INCA_IP2_GPIO_ALTSEL1 = 0x0000003c;
#endif
	/* Set DIR for ASC0 and ASC1 (TX=Out) */
	*INCA_IP2_GPIO_DIR |= 0x00820000;

	/* and we have to set CLC register*/
	pAsc->asc_clc = 0x00000100;

	/* initialy we are in async mode */
	pAsc->asc_mcon = ASCMCON_M_8ASYNC | 0x00c00000 ;

	/* select input port */
	pAsc->asc_pisel = (CONSOLE_TTY & 0x1);

	/* TXFIFO's filling level */
	/* enable TXFIFO */
	pAsc->asc_txfcon |= ASCTXFCON_TXFEN | ASCTXFCON_TXFIFO1;

	/* RXFIFO's filling level */
	/* enable RXFIFO */
	//	pAsc->asc_rxfcon |= ASCRXFCON_RXFEN | ASCRXFCON_RXFIFO1;
	pAsc->asc_rxfcon |= ASCRXFCON_RXFIFO1;

	/* enable error signals */
	reg = pAsc->asc_mcon;
	reg = reg | ASCMCON_FEN;
	pAsc->asc_mcon = reg;

	/* Clear all error interrupts and disable receiver */
	pAsc->asc_whbstate=0x000000fd;

	/* acknowledge ASC interrupts */
	ASC_INTERRUPTS_CLEAR(INCA2ASC_IRQ_LINE_ALL);

	/* disable ASC interrupts */
	ASC_INTERRUPTS_DISABLE(INCA2ASC_IRQ_LINE_ALL);

	/* set baud rate */
	serial_setbrg();

	if ((u32) pAsc == (u32) INCA_IP2_ASC0) {
		/* Set FIFO to single byte forward mode */
		pAsc->asc_eomcon = 0x00010300;
	}
   
	/* Enable receiver */
	pAsc->asc_whbstate=0x00000002;

	/* and we have to set CLC register*/
//	CLEAR_BIT(pAsc->asc_clc, ASCCLC_DISS);
//	SET_BITFIELD(pAsc->asc_clc, ASCCLC_RMCMASK, ASCCLC_RMCOFFSET, 0x0001);

	/* initialy we are in async mode */
//	pAsc->asc_mcon = ASCMCON_M_8ASYNC;

	/* select input port */
//	pAsc->asc_pisel = (CONSOLE_TTY & 0x1);

	/* TXFIFO's filling level */
//	SET_BITFIELD(pAsc->asc_txfcon, ASCTXFCON_TXFITLMASK,
//	             ASCTXFCON_TXFITLOFF, INCA2ASC_TXFIFO_FL);
	/* enable TXFIFO */
//	SET_BIT(pAsc->asc_txfcon, ASCTXFCON_TXFEN);

	/* RXFIFO's filling level */
//	SET_BITFIELD(pAsc->asc_txfcon, ASCRXFCON_RXFITLMASK,
//	             ASCRXFCON_RXFITLOFF, INCA2ASC_RXFIFO_FL);
	/* enable RXFIFO */
//	SET_BIT(pAsc->asc_rxfcon, ASCRXFCON_RXFEN);

	/* enable error signals */
//	SET_BIT(pAsc->asc_mcon, ASCMCON_FEN);

//	pAsc->asc_whbstate=0x000000fffd;

	/* acknowledge ASC interrupts */
//	ASC_INTERRUPTS_CLEAR(INCA2ASC_IRQ_LINE_ALL);

	/* disable ASC interrupts */
//	ASC_INTERRUPTS_DISABLE(INCA2ASC_IRQ_LINE_ALL);

	/* set baud rate */
//	serial_setbrg();

	/* set the options */
//	serial_setopt();

	/* Enable receiver */
//	pAsc->asc_whbstate=0x00000002;

	return 0;
#endif /* EMULATION */
}

void serial_setbrg (void)
{
#ifdef EMULATION

	u32 reg;

	/* Disable Baud Rate Generator; BG should only be written when R=0 */
	reg = pAsc->asc_mcon;
	reg = reg & ~ASCMCON_R;
	reg = reg | ASCMCON_FDE;
	pAsc->asc_mcon = reg;

#ifdef VENUS
   /* Set fractional divider and reload value for 19200 baud */
   /* fVENUS = 10MHz */
   pAsc->asc_fdv = 472;
   pAsc->asc_bg  = 4;
#else
	/* Check for Emulator */
	if ((*INCA_IP2_GPIO_IN & 0x00002000) != 0) {
		/* PRO */
		/* Set fractional divider value for 600@74kHz */
		pAsc->asc_fdv = 465;
		/* Set reload value in BG for 600@74kHz */
		pAsc->asc_bg = 2;
	}
	else {
		/* Set fractional divider value for 600baud, div 220 */
		pAsc->asc_fdv = 382;
		/* Set reload value in BG */
		pAsc->asc_bg = 1;
	}
#endif

	/* Enable Baud Rate Generator */
	reg = pAsc->asc_mcon;
	reg = reg | ASCMCON_R | ASCMCON_FDE | ASCMCON_FEN;
	pAsc->asc_mcon = reg;

#else

	ulong      uiReloadValue, fdv;

#if 0
	u32 reg;

    /* Set fractional divider and reload value for 115200 baud */
   /* freq = 396MHz (with DDR-SDRAM results in 133MHz for ASC) */
#ifdef INCAIP2_USE_DDR_RAM
    pAsc->asc_fdv = 360;
    pAsc->asc_bg  = 24;
#else
    /* freq = 396MHz (with SDRAM results in 100MHz for ASC) */
    pAsc->asc_fdv = 365;
    pAsc->asc_bg  = 18;
#endif

	/* Enable Baud Rate Generator */
	reg = pAsc->asc_mcon;
	reg = reg | ASCMCON_R | ASCMCON_FDE | ASCMCON_FEN;
	pAsc->asc_mcon = reg;

   return;

#else
	ulong      f_ASC;

	f_ASC = incaip2_get_fbs1clk();

	fdv = INCA2ASC_FDV_HIGH_BAUDRATE;
	uiReloadValue = (f_ASC / (8192 * CONFIG_BAUDRATE / fdv)) - 1;

	if ( (uiReloadValue < 0) || (uiReloadValue > 8191) )
	{
		fdv = INCA2ASC_FDV_LOW_BAUDRATE;
		uiReloadValue = (f_ASC / (8192 * CONFIG_BAUDRATE / fdv)) - 1;

		if ( (uiReloadValue < 0) || (uiReloadValue > 8191) )
		{
			return;    /* can't impossibly generate that baud rate */
		}
	}

	/* Disable Baud Rate Generator; BG should only be written when R=0 */
	CLEAR_BIT(pAsc->asc_mcon, ASCMCON_R);


	/* Enable Fractional Divider */
	SET_BIT(pAsc->asc_mcon, ASCMCON_FDE); /* FDE = 1 */

	/* Set fractional divider value */
	pAsc->asc_fdv = fdv & ASCFDV_VALUE_MASK;

	/* Set reload value in BG */
	pAsc->asc_bg = uiReloadValue;

	/* Enable Baud Rate Generator */
	SET_BIT(pAsc->asc_mcon, ASCMCON_R);           /* R = 1 */
#endif /* 1 */
#endif /* EMULATION */
}

/*******************************************************************************
*
* serial_setopt - set the serial options
*
* Set the channel operating mode to that specified. Following options
* are supported: CREAD, CSIZE, PARENB, and PARODD.
*
* Note, this routine disables the transmitter.  The calling routine
* may have to re-enable it.
*
* RETURNS:
* Returns 0 to indicate success, otherwise -1 is returned
*/

static int serial_setopt (void)
{
	ulong  con;

	return 0;

	switch ( ASC_OPTIONS & ASCOPT_CSIZE )
	{
		/* 7-bit-data */
	case ASCOPT_CS7:
		con = ASCMCON_M_7ASYNC | ASCMCON_PEN;   /* 7-bit-data and parity bit */
		break;

		/* 8-bit-data */
	case ASCOPT_CS8:
		con = ASCMCON_M_8ASYNC;   /* 8-bit-data */
		if ( ASC_OPTIONS & ASCOPT_PARENB )
			con |= ASCMCON_PEN;   /* 8-bit-data and parity bit */
		break;

		/*
		 *  only 7 and 8-bit frames are supported
		 *  if we don't use IOCTL extensions
		 */
	default:
		return -1;
	}

	if ( ASC_OPTIONS & ASCOPT_STOPB )
		SET_BIT(con, ASCMCON_STP);       /* 2 stop bits */
	else
		CLEAR_BIT(con, ASCMCON_STP);     /* 1 stop bit */

	if ( ASC_OPTIONS & ASCOPT_PARENB )
		SET_BIT(con, ASCMCON_PEN);           /* enable parity checking */
	else
		CLEAR_BIT(con, ASCMCON_PEN);         /* disable parity checking */

	if ( ASC_OPTIONS & ASCOPT_PARODD )
		SET_BIT(con, ASCMCON_ODD);       /* odd parity */
	else
		CLEAR_BIT(con, ASCMCON_ODD);     /* even parity */

	if ( ASC_OPTIONS & ASCOPT_CREAD )
		SET_BIT(pAsc->asc_whbstate, ASCWHBSTATE_SETREN); /* Receiver enable */

	pAsc->asc_mcon |= con;

	return 0;

}

void serial_putc (const char c)
{
#ifdef EMULATION
	u32 txFl = 0;

	if (c == '\n') serial_putc ('\r');

	/* check for free space in the TX FIFO */
	do
	{
		txFl = ( pAsc->asc_fstat & ASCFSTAT_TXFREEMASK ) >> ASCFSTAT_TXFREEOFF;
	}
	while ( txFl == 0 );

	if ((u32) pAsc == (u32) INCA_IP2_ASC0) {
		*(((char*) &pAsc->asc_tbuf)+3) = c; /* write char to Transmit Buffer Register */
	}
	else {
		pAsc->asc_tbuf = c; /* write char to Transmit Buffer Register */
	}
	asm("sync");

	/* check for errors */
	if ( pAsc->asc_state & ASCSTATE_TOE )
	{
		SET_BIT(pAsc->asc_whbstate, ASCWHBSTATE_CLRTOE);
		return;
	}
#else

	uint txFl = 0;
#if 0
    /********************************************
	 * JUST FOR SIMULATION
	 ********************************************/

	char* p=(char*) MON_PUTCH;

	if (c == '\n') serial_putc ('\r');
	*p=c;
	return;
#endif

	if (c == '\n') serial_putc ('\r');

	/* check for free space in the TX FIFO */
	do
	{
		txFl = ( pAsc->asc_fstat & ASCFSTAT_TXFREEMASK ) >> ASCFSTAT_TXFREEOFF;
	}
	while ( txFl == 0 );

	if ((u32) pAsc == (u32) INCA_IP2_ASC0) {
#ifdef __LITTLE_ENDIAN
		*((char*) &pAsc->asc_tbuf) = c; /* write char to Transmit Buffer Register */
#else /* __BIG_ENDIAN */
		*(((char*) &pAsc->asc_tbuf)+3) = c; /* write char to Transmit Buffer Register */
#endif
	}
	else {
		pAsc->asc_tbuf = c; /* write char to Transmit Buffer Register */
	}
	asm("sync");

	/* check for errors */
	if ( pAsc->asc_state & ASCSTATE_TOE )
	{
		SET_BIT(pAsc->asc_whbstate, ASCWHBSTATE_CLRTOE);
		return;
	}
#endif /* EMULATION */
}

void serial_puts (const char *s)
{
	while (*(s))
	{
		serial_putc(*s++);
	}
}

int serial_getc (void)
{
#ifdef EMULATION
	ulong symbol_mask;
	char c;

	while (!serial_tstc());

	symbol_mask =
	     ((ASC_OPTIONS & ASCOPT_CSIZE) == ASCOPT_CS7) ? (0x7f) : (0xff);
	
	c = (char)(pAsc->asc_rbuf & symbol_mask);

	return c;
#else
	ulong symbol_mask;
	char c;

#if 0
    /********************************************
	 * JUST FOR SIMULATION
	 ********************************************/
	return '.';
#endif

	while (!serial_tstc());

	symbol_mask =
	     ((ASC_OPTIONS & ASCOPT_CSIZE) == ASCOPT_CS7) ? (0x7f) : (0xff);
	
	c = (char)(pAsc->asc_rbuf & symbol_mask);

	return c;

#endif /* EMULATION */
}

/* Test for char on receive channel and and clear errors */
int serial_tstc (void)
{
	int res = 1;

	if ( (pAsc->asc_fstat & ASCFSTAT_RXFFLMASK) == 0 )
	{
		res = 0;
	}
	else if ( pAsc->asc_state & ASCSTATE_FE )
	{
		SET_BIT(pAsc->asc_whbstate, ASCWHBSTATE_CLRFE);
		res = 0;
	}
	else if ( pAsc->asc_state & ASCSTATE_PE )
	{
		SET_BIT(pAsc->asc_whbstate, ASCWHBSTATE_CLRPE);
		res = 0;
	}
	else if ( pAsc->asc_state & ASCSTATE_ROE )
	{
		SET_BIT(pAsc->asc_whbstate, ASCWHBSTATE_CLRROE);
		res = 0;
	}

	return res;
}

#ifdef CONFIG_INCA_IP2_BLUETOOTH_DISABLE

#define INCAIP2_DISABLE_BLUETOOTH_DEBUG

#define INCAIP2_DISABLE_BLUETOOTH_RESET_LENGTH 4
#define INCAIP2_DISABLE_BLUETOOTH_LPM_LENGTH 6
#define INCAIP2_DISABLE_BLUETOOTH_MODE_LENGTH 6
#define INCAIP2_DISABLE_BLUETOOTH_4WIRE_LENGTH 62
#define INCAIP2_DISABLE_BLUETOOTH_READ_DATA_LENGTH 12
static volatile inca2Asc_t *pAscBT = (inca2Asc_t *)INCA_IP2_ASC0;
char incaip2_disable_bluetooth_mode[] = { 
					0x01, 0x11, 0xFC, 0x02, 0x01, 0x00
					};
char incaip2_disable_bluetooth_4wire[] = { 
					0x01, 0x2F, 0xFC, 0x3A, 0x00, 0x8B, 
					0x9E, 0x19, 0x03, 0x00, 0x00, 0x00,
					0x80, 0x00, 0x80, 0xBA, 0x8C, 0x01,
					0xEF, 0xFF, 0x8D, 0xFE, 0x9B, 0x19,
					0x00, 0x01, 0x2F, 0xFC, 0x3A, 0x00,
					0x8B, 0x9E, 0x19, 0x03, 0x00, 0x00,
					0x00, 0x80, 0x00, 0x90, 0x04, 0x8C,
					0x01, 0xEF, 0xFF, 0x8D, 0xFE, 0x9B,
					0x19, 0x00, 0x00, 0x00, 0x03, 0x00,
					0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
					0x00, 0x00
					};

int incaip2_disable_bluetooth(void)
{
	int t=0;
	int i=0;
	int num_chars1=0;
	int num_chars2=0;
	int rx_count=0;
	char rx_buffer[255];
	char c;
	ulong symbol_mask;
	u32 reg;
	
	reg = *INCA_IP2_GPIO_ALTSEL2;
	reg &= ~0x000c0000;
	reg |= 0x00040000;
	*INCA_IP2_GPIO_ALTSEL2 = reg;
	*INCA_IP2_GPIO_DIR |= 0x00000200;
	*INCA_IP2_GPIO_OUT |= 0x00000200;
	
	
	i=0;
	while ((i+rx_count) < 255) {
		t=0;
		while ((!serial_tstc()) && (t < 700000)) t++;
		if (t < 700000) {
			num_chars1++;
			symbol_mask =
				((ASC_OPTIONS & ASCOPT_CSIZE) == ASCOPT_CS7) ? (0x7f) : (0xff);
			rx_buffer[rx_count+i] = (char)(pAsc->asc_rbuf & symbol_mask);
			
		}
		else break;
		i++;
	}
	rx_count+=i;
	
	for (t=0; t<INCAIP2_DISABLE_BLUETOOTH_MODE_LENGTH ; t++)
		serial_putc(incaip2_disable_bluetooth_mode[t]);
	
	i=0;
	while ((i+rx_count) < 255) {
		t=0;
		while ((!serial_tstc()) && (t < 700000)) t++;
		if (t < 700000) {
			num_chars1++;
			symbol_mask =
				((ASC_OPTIONS & ASCOPT_CSIZE) == ASCOPT_CS7) ? (0x7f) : (0xff);
			rx_buffer[rx_count+i] = (char)(pAsc->asc_rbuf & symbol_mask);
			
		}
		else break;
		i++;
	}
	rx_count+=i;
	
	for (t=0; t<INCAIP2_DISABLE_BLUETOOTH_4WIRE_LENGTH ; t++)
		serial_putc(incaip2_disable_bluetooth_4wire[t]);
	i=0;
	while ((i+rx_count) < 255) {
		t=0;
		while ((!serial_tstc()) && (t < 700000)) t++;
		if (t < 700000) {
			num_chars2++;
			symbol_mask =
				((ASC_OPTIONS & ASCOPT_CSIZE) == ASCOPT_CS7) ? (0x7f) : (0xff);
			rx_buffer[rx_count+i] = (char)(pAsc->asc_rbuf & symbol_mask);
		}
		else break;
		i++;
	}
	
#ifdef INCAIP2_DISABLE_BLUETOOTH_DEBUG
	serial_puts("\n### Sent Bluemoon disable sequence\n\0");	
#endif	
	serial_puts("\n### Received buffer:\n\0");
	for (t=0;t<rx_count;t++) {
		if (t % 8 == 0) printf("\n");
		printf("0x%02x ",rx_buffer[t]&0xff);	
	}
	printf("\n");
	return 0;
}
#endif
