/*
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 *
 *  ifx_nand.c ifx_nand.h
 *
 *  Description:
 *    device driver of NAND controller of INCAIP2/Danube chip
 *  Author:
 *    Samuels Xu Liang
 *  Created:
 *    6 Sep 2005
 *  History & Modification Tag:
 * ########################################################################
 *
 *  ___________________________________________________________________________
 *  |  Tag   |                  Comments                   | Modifier & Time  |
 *  |--------+---------------------------------------------+------------------|
 *  |  S0.0  | First version of this driver and the tag is | Samuels Xu Liang |
 *  |        | implied.                                    |    6 Jul 2005    |
 *  ---------------------------------------------------------------------------
 *
 *  ___________________________________________________________________________
 *  |  Tag   |                  Comments                   | Modifier & Time  |
 *  |--------+---------------------------------------------+------------------|
 *  |  S0.1  | Second version of this driver for INCAIP2   | Winder Sung      |
 *  |        |                                             |    3 Nov 2005    |
 *  ---------------------------------------------------------------------------
 */
/*
 *  Common Head File
 */
#include <asm/addrspace.h>

#ifdef NAND_DEBUG
#define NANDPRINT printf
#else
#define NANDPRINT(fmt,args...)
#endif
 /*
 *  Chip Specific Head File
 */
//#define IFX_EBU_ADDR_REGION0 0x10000000 /* Default CS0 used by EBU for external boot after reset. */
//#define IFX_EBU_ADDR_REGION1 0x19000000 /* CS1 */
//#define IFX_EBU_ADDR_REGION2 0x1B000000 /* PC card */
//#define IFX_EBU_ADDR_REGION3 0x1C000000
//#define IFX_EBU_ADDR_REGION_BASE IFX_EBU_ADDR_REGION0 /* NAND flash in region0 (ebu_address_region  = 0) */
#define IFX_EBU_REGISTER_BASE 0x14102000

/*
 * ####################################
 *     Parameters to Configure NAND
 * ####################################
 */
static volatile u8 *ebu_base_address = (volatile u8*)(CFG_NAND_BASE);
#if (CFG_NAND_BASE == (KSEG1 | IFX_EBU_ADDR_REGION0))
#define ebu_address_region          0
#elif (CFG_NAND_BASE == (KSEG1 | IFX_EBU_ADDR_REGION1))
#define ebu_address_region          1
#elif (CFG_NAND_BASE == (KSEG1 | IFX_EBU_ADDR_REGION2))
#define ebu_address_region          2
#elif (CFG_NAND_BASE == (KSEG1 | IFX_EBU_ADDR_REGION3))
#define ebu_address_region          3
#endif

/*
 * ####################################
 *              Definition
 * ####################################
 */
/*
 *  Bits Operation
 */
#define GET_BITS(x, msb, lsb)           (((x) & ((1 << ((msb) + 1)) - 1)) >> (lsb))
#define SET_BITS(x, msb, lsb, value)    (((x) & ~(((1 << ((msb) + 1)) - 1) ^ ((1 << (lsb)) - 1))) | (((value) & ((1 << (1 + (msb) - (lsb))) - 1)) << (lsb)))

/*
 *  EBU Registers Mapping
 */
#define IFX_EBU_REG                  (KSEG1 | IFX_EBU_REGISTER_BASE)

#define IFX_EBU_ADDR_SEL             ((volatile u32*)(IFX_EBU_REG + 0x20 + (ebu_address_region * 4)))
#define IFX_EBU_CONFIG               ((volatile u32*)(IFX_EBU_REG + 0x60 + (ebu_address_region * 4)))

/*
 *  NAND Registers Mapping
 */
#define IFX_NAND_REG                 IFX_EBU_REG

#define IFX_NAND_CONTROL             ((volatile u32*)(IFX_NAND_REG + 0x00B0))
#define IFX_NAND_WAIT                ((volatile u32*)(IFX_NAND_REG + 0x00B4))
#define IFX_NAND_ECC0                ((volatile u32*)(IFX_NAND_REG + 0x00B8))
#define IFX_NAND_ECC_AC              ((volatile u32*)(IFX_NAND_REG + 0x00BC))

/*
 *  NAND IO Port Mapping
 */
#define NAND_CMD_DECODER_ALE(x)      ((x) ? (1 << 2) : 0)
#define NAND_CMD_DECODER_CLE(x)      ((x) ? (1 << 3) : 0)
#define NAND_CMD_DECODER_CS(x)       ((x) ? (1 << 4) : 0)
#define NAND_CMD_DECODER_SE(x)       ((x) ? (1 << 5) : 0)
#define NAND_CMD_DECODER_WP(x)       ((x) ? (1 << 6) : 0)
#define NAND_CMD_DECODER_PRE(x)      ((x) ? (1 << 7) : 0)

#define IFX_NAND_CMD                 ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(1) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_WRITE               ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_ADDR                ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(1) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_READ                ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))

#define IFX_NAND_WRITE_OVER          ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_ADDR_OVER           ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_READ_OVER           ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))

/*
 *  EBU Address Select Register
 */
#define EBU_ADDR_SEL_BASE_SET(value)    (((value) & ((1 << (1 + 31 - 12)) - 1)) << 12)
#define EBU_ADDR_SEL_MASK_SET(value)    SET_BITS(0, 7, 4, (value))
#define EBU_ADDR_SEL_MRME_SET(value)    ((value) ? (1 << 1) : 0)
#define EBU_ADDR_SEL_REGEN_SET(value)   ((value) ? (1 << 0) : 0)

/*
 *  EBU Configuration Register
 */
#define EBU_CON_WRDIS_SET(value)        ((value) ? (1 << 31) : 0)
#define EBU_CON_ADSWP_SET(value)        ((value) ? (1 << 30) : 0)
#define EBU_CON_AGEN_SET(value)         SET_BITS(0, 26, 24, (value))
#define EBU_CON_SETUP_SET(value)        ((value) ? (1 << 22) : 0)
#define EBU_CON_WAIT_SET(value)         SET_BITS(0, 21, 20, (value))
#define EBU_CON_WINV_SET(value)         ((value) ? (1 << 19) : 0)
#define EBU_CON_PW_SET(value)           SET_BITS(0, 17, 16, (value))
#define EBU_CON_ALEC_SET(value)         SET_BITS(0, 15, 14, (value))
#define EBU_CON_BCGEN_SET(value)        SET_BITS(0, 13, 12, (value))
#define EBU_CON_WAITWRC_SET(value)      SET_BITS(0, 10, 8, (value))
#define EBU_CON_WAITRDC_SET(value)      SET_BITS(0, 7, 6, (value))
#define EBU_CON_HOLDC_SET(value)        SET_BITS(0, 5, 4, (value))
#define EBU_CON_RECOVC_SET(value)       SET_BITS(0, 3, 2, (value))
#define EBU_CON_CMULT_SET(value)        SET_BITS(0, 1, 0, (value))

/*
 *  NAND Flash Controller Control Register
 */
#define NAND_CON_LAT_EN                 GET_BITS(*IFX_NAND_CONTROL, 23, 18)

#define NAND_CON_ECC_ON_SET(value)      ((value) ? (1 << 31) : 0)
#define NAND_CON_PRE_LAT_EN_SET(value)  ((value) ? (1 << 23) : 0)
#define NAND_CON_WP_LAT_EN_SET(value)   ((value) ? (1 << 22) : 0)
#define NAND_CON_SE_LAT_EN_SET(value)   ((value) ? (1 << 21) : 0)
#define NAND_CON_CS_LAT_EN_SET(value)   ((value) ? (1 << 20) : 0)
#define NAND_CON_CLE_LAT_EN_SET(value)  ((value) ? (1 << 19) : 0)
#define NAND_CON_ALE_LAT_EN_SET(value)  ((value) ? (1 << 18) : 0)
#define NAND_CON_OUT_CS_S_SET(value)    SET_BITS(0, 11, 10, (value))
#define NAND_CON_IN_CS_S_SET(value)     SET_BITS(0, 9, 8, (value))
#define NAND_CON_PRE_P_SET(value)       ((value) ? (1 << 7) : 0)
#define NAND_CON_WP_P_SET(value)        ((value) ? (1 << 6) : 0)
#define NAND_CON_SE_P_SET(value)        ((value) ? (1 << 5) : 0)
#define NAND_CON_CS_P_SET(value)        ((value) ? (1 << 4) : 0)
#define NAND_CON_CLE_P_SET(value)       ((value) ? (1 << 3) : 0)
#define NAND_CON_ALE_P_SET(value)       ((value) ? (1 << 2) : 0)
#define NAND_CON_CSMUX_E_SET(value)     ((value) ? (1 << 1) : 0)
#define NAND_CON_NANDM_SET(value)       ((value) ? (1 << 0) : 0)

/*
 *  NAND Flash Device RD/BY State Register
 */
#define NAND_WAIT_WR_C          (*IFX_NAND_WAIT & (1 << 3))
#ifdef NAND_INVERT_RB
#define NAND_WAIT_RYBY          ((*IFX_NAND_WAIT & 0x7) == 0x6)
#else
#define NAND_WAIT_RYBY          ((*IFX_NAND_WAIT & 0x7) == 0x7)
#endif
#define NAND_WAIT_CLEAR_ALL()   do{*IFX_NAND_WAIT = 0x0;}while(0)

/*
 *  NAND Flash ECC Register
 */
#define NAND_ECC0_B2                    GET_BITS(*IFX_NAND_ECC0, 23, 16)
#define NAND_ECC0_B1                    GET_BITS(*IFX_NAND_ECC0, 15, 8)
#define NAND_ECC0_B0                    GET_BITS(*IFX_NAND_ECC0, 7, 0)

#define NAND_ECC0_CLEAR()               do { *IFX_NAND_ECC0 = 0; *IFX_NAND_ECC_AC = 0; } while (0)

/*
 * ####################################
 *             Declaration
 * ####################################
 */
// select chip
#define NAND_ENABLE_CE(nand) do{*IFX_NAND_CONTROL |= NAND_CON_CS_LAT_EN_SET(1);}while(0)
// de-select chip
#define NAND_DISABLE_CE(nand) do{*IFX_NAND_CONTROL &= ~NAND_CON_CS_LAT_EN_SET(1);}while(0)
// Assert the CLE (Command Latch Enable) line to the flash chip
#define NAND_CTL_SETCLE(nandptr)
// Lower the CLE line
#define NAND_CTL_CLRCLE(nandptr)
// Assert the ALE (Adress Latch Enable) line to the flash chip
#define NAND_CTL_SETALE(nandptr)
// Lower the ALE line
#define NAND_CTL_CLRALE(nandptr)
// Write Command
#define WRITE_NAND_COMMAND(command, nandptr) do{*IFX_NAND_CMD = command;NANDPRINT("%s: CMD:0x%x, Value:0x%x\n", __FUNCTION__,(u32)IFX_NAND_CMD, command);while(!NAND_WAIT_WR_C);}while(0)
// Write Address
#define WRITE_NAND_ADDRESS(ofs, nandptr) do{*IFX_NAND_ADDR = ofs;NANDPRINT("%s: ADD:0x%x, Value:0x%x\n", __FUNCTION__,(u32)IFX_NAND_ADDR, (u32)ofs);while(!NAND_WAIT_WR_C);}while(0)

#define WRITE_NAND_ADDRESS_OVER(ofs, nandptr) do{*IFX_NAND_ADDR_OVER = ofs;NANDPRINT("%s: ADD:0x%x, Value:0x%x\n", __FUNCTION__,(u32)IFX_NAND_ADDR_OVER, (u32)ofs);while(!NAND_WAIT_WR_C);}while(0)
//Write Data
#define WRITE_NAND(d, adr) do {*(volatile unsigned char *)((unsigned long)(IFX_NAND_WRITE)) = (unsigned char)(d);while(!NAND_WAIT_WR_C);} while(0)
// Read Date
#define READ_NAND(adr) (*(volatile unsigned char *)(unsigned long)(IFX_NAND_READ))

// Wait device ready, write complete
#define NAND_WAIT_READY(nand) do{int wait=0x40000;NANDPRINT("Wait for Ready!!\n");while(!NAND_WAIT_RYBY && wait--);NAND_WAIT_CLEAR_ALL();}while(0)

