#include <common.h>
#include <asm/inca-ip.h>

#if (CONFIG_COMMANDS && CFG_CMD_NAND)

#include "incaip2_nand.h"
#include <linux/mtd/nand.h>

#ifdef CONFIG_MTD_NAND_HW_ECC
/* The following header is included to access the generic
 * nand_correct_data function, which is used now as the 
 * correct data function for hw ecc */
#include <linux/mtd/nand_ecc.h>
#endif

/* Hardware control function */
static void incaip_hwcontrol (struct mtd_info *mtd, int cmd)
{
	struct nand_chip *this = (struct nand_chip *) mtd->priv;

	switch (cmd) {
		case NAND_CTL_SETCLE:
			(unsigned long) this->IO_ADDR_W |= NAND_CLE_BIT;
			break;
		case NAND_CTL_CLRCLE:
			(unsigned long) this->IO_ADDR_W &= ~NAND_CLE_BIT;
			break;
		case NAND_CTL_SETALE:
			(unsigned long) this->IO_ADDR_W |= NAND_ALE_BIT;
			break;
		case NAND_CTL_CLRALE:
			(unsigned long) this->IO_ADDR_W &= ~NAND_ALE_BIT;
			break;
		case NAND_CTL_SETNCE:
			*IFX_NAND_CONTROL |= NAND_CON_NANDM_SET(1);
			break;
		case NAND_CTL_CLRNCE:
			*IFX_NAND_CONTROL &= ~NAND_CON_NANDM_SET(1);
			break;
	}
}

/* Device ready function */
int incaip_dev_ready (struct mtd_info *mtd)
{
	return ( NAND_WAIT_RD ? 1 : 0);
}

#ifdef CONFIG_MTD_NAND_HW_ECC
int incaip_calculate_ecc (struct mtd_info *mtd, const u_char *dat, u_char *ecc_code)
{
	u32 ecc = 0;
	/* IFX_NAND_ECC0: Read the ECC bits */
	/* Byte read reordered: Danube ECC bit order */
	
	ecc = *IFX_NAND_ECC0;
	NAND_ECC0_CLEAR();
	
	ecc = (~(ecc << 2)) & 0xffffff;
	
	ecc_code[2] = ecc & 0xFF;
	ecc_code[1] = (ecc >> 8) & 0xFF;
	ecc_code[0] = (ecc >> 16) & 0xFF;

	return 0;
}

void incaip_enable_hwecc (struct mtd_info *mtd, int mode)
{
	struct nand_chip *this = mtd->priv;
	
	switch(mode) {
	case NAND_ECC_READ:
		this->calculate_ecc = incaip_calculate_ecc;
		break;
	case NAND_ECC_WRITE:
		this->calculate_ecc = nand_calculate_ecc;
		break;
	}

	*IFX_NAND_CONTROL |= NAND_CON_ECC_ON_SET(1);
	NAND_ECC0_CLEAR();
	*IFX_NAND_ECC_AC = 0;
}
#endif

/* Initialize hardware to support NAND device */
void incaip_hwinit (void) 
{
    *IFX_EBU_ADDR_SEL = EBU_ADDR_SEL_BASE_SET(PHYSADDR(CFG_NAND_BASE) >> 12)
                         | EBU_ADDR_SEL_MASK_SET(15)
                         | EBU_ADDR_SEL_MRME_SET(0)
                         | EBU_ADDR_SEL_REGEN_SET(1);

    *IFX_EBU_CONFIG = EBU_CON_WRDIS_SET(0) | EBU_CON_ADSWP_SET(1) | EBU_CON_AGEN_SET(0x00)
                    | EBU_CON_SETUP_SET(1) | EBU_CON_WAIT_SET(0x00) | EBU_CON_WINV_SET(0)
                    | EBU_CON_PW_SET(0x00) | EBU_CON_ALEC_SET(3) | EBU_CON_BCGEN_SET(0x01)
                    | EBU_CON_WAITWRC_SET(2) | EBU_CON_WAITRDC_SET(2) | EBU_CON_HOLDC_SET(1)
                    | EBU_CON_RECOVC_SET(1) | EBU_CON_CMULT_SET(0x01);

    switch (ebu_address_region) {
        case 2:
            *INCA_IP2_GPIO_ALTSEL1 |= 0x00003000;
            *INCA_IP2_GPIO_DIR |= 0x00400000;
            break;
        case 3:
            *INCA_IP2_GPIO_ALTSEL1 |= 0x00000C00;
            *INCA_IP2_GPIO_DIR |= 0x00200000;
            break;
    }
	
    /*
     *  ECC disabled
     *  CLE, ALE and CS are pulse, all other signal are latches based
     *  CLE and ALE are active high, PRE, WP, SE and CS/CE are active low
     *  OUT_CS_S is disabled
     *  NAND mode is disabled
     */
    *IFX_NAND_CONTROL = NAND_CON_ECC_ON_SET(0)
                     | NAND_CON_PRE_LAT_EN_SET(1) | NAND_CON_WP_LAT_EN_SET(1)
                     | NAND_CON_SE_LAT_EN_SET(1) | NAND_CON_CS_LAT_EN_SET(0)
                     | NAND_CON_CLE_LAT_EN_SET(0) | NAND_CON_ALE_LAT_EN_SET(0)
                     | NAND_CON_OUT_CS_S_SET(ebu_address_region)
                     | NAND_CON_IN_CS_S_SET(ebu_address_region)
                     | NAND_CON_PRE_P_SET(1) | NAND_CON_WP_P_SET(1)
                     | NAND_CON_SE_P_SET(1) | NAND_CON_CS_P_SET(1)
                     | NAND_CON_CLE_P_SET(0) | NAND_CON_ALE_P_SET(0)
                     | NAND_CON_CSMUX_E_SET(0)
                     | NAND_CON_NANDM_SET(0);
}

/* Board initialization function */
void board_nand_init (struct nand_chip *nand) 
{
	
	/* Set hardware control function */
	nand->hwcontrol = incaip_hwcontrol;

	/* Assign the device ready function */
	nand->dev_ready = incaip_dev_ready;

#ifdef CONFIG_MTD_NAND_HW_ECC
	nand->eccmode = NAND_ECC_HW3_256;
	nand->enable_hwecc = incaip_enable_hwecc;
	nand->calculate_ecc = incaip_calculate_ecc;
	nand->correct_data = nand_correct_data;
#else
	nand->eccmode = NAND_ECC_SOFT;
#endif

	/* On the incaip2, the IO_ADDR to read and write is not the base address, but as follows */
	(unsigned long) nand->IO_ADDR_W |= NAND_CMD_DECODER_CS(1);
	(unsigned long) nand->IO_ADDR_R |= NAND_CMD_DECODER_CS(1);

	/* Initialize relevant ports and the EBU register */
	incaip_hwinit();

	/* Buffers are allocated in the nand_scan function */
	return;
}

#endif /* CFG_CMD_NAND */
