/*****************************************************************************
 * INCA-IP2 BootROM
 * Copyright (c) 2005, Infineon Technologies AG, All rights reserved
 * COM NRW CSE VoIP
 *****************************************************************************/
#include <common.h>

#if defined(CONFIG_SERIAL_FLASH) || defined(CONFIG_DISPLAY)

#include "ssc.h"
#include <asm/inca-ip2.h>
#include <asm/addrspace.h>


int ssc_init(void)
{

#if (SSC_MODULE == 0)
   /* we have to set PMS.GEN13 bit to enable an ASC0 device*/
   *INCA_IP2_PMS_PMS_GEN |= INCA_IP2_PMS_PMS_GEN_SPI0;

   /* Enable SPI0 */
   *INCA_IP2_GPIO_ALTSEL1 |= 0x0000002a;
   *INCA_IP2_GPIO_ALTSEL2 |= 0x80000000;

   /* Set DIR for SPI0 */
   *INCA_IP2_GPIO_DIR    |= 0x00038000;
#else
   /* we have to set PMS.GEN13 bit to enable an ASC0 device*/
   *INCA_IP2_PMS_PMS_GEN |= INCA_IP2_PMS_PMS_GEN_SPI1;
   
   /* Enable SPI1 */
   *INCA_IP2_GPIO_ALTSEL2 |= 0x0000aa00;

   /* Set DIR for SPI1 */
   *INCA_IP2_GPIO_DIR    |= 0x00000070;
#endif

    /* Disable SSC */
    *SSC_WHBSTATE = 0x1;
    *SSC_CLC = 0x4000;
    asm("SYNC");
    
    /* GPIO CS */
    *SSC_GPOCON     = 0x0100;
    *SSC_WHBGPOSTAT = 0x0100; //CS to high
    /* disable IRQ */
    *SSC_IRN_EN = 0x0;
    /* set Baudrate */
    *SSC_BR = 0x02; /* (fpi clock/2)/(BR + 1) */

    /* Set Master mode */
    *SSC_WHBSTATE =0x8;
    /* Enable and flush RX/TX FIFO */
    *SSC_RXFCON = 0xF03;
    *SSC_TXFCON = 0xF03;

    *SSC_CON = 0x00070030;
    asm("SYNC");    
        
    /* Enable SSC */
    *SSC_WHBSTATE = 0x2;
    asm("SYNC");

    return 0;
}

int ssc_read(unsigned char *uc_value)
{
    while(!((*SSC_FSTAT) & 0x3F));
    *uc_value = ((*SSC_RB) & 0xFF);
    return 0;
}


int ssc_write(unsigned char uc_value)
{
    while((*SSC_STATE) & 0x2000);
    *SSC_TB = uc_value;
    asm("sync");
    return 0;
}

int ssc_transceive(unsigned char out, unsigned char *in)
{
    char tmp;

    if (in == NULL)
        in = &tmp;
    while((*SSC_STATE) & 0x2000);
    *SSC_TB = out;
    asm("sync");
    while(!((*SSC_FSTAT) & 0x3F));
    *in = ((*SSC_RB) & 0xFF);
    return 0;
}

int ssc_close(void)
{
    *SSC_CLC = 0x00000002;

#if (SSC_MODULE == 0)
    *INCA_IP2_PMS_PMS_GEN &= ~INCA_IP2_PMS_PMS_GEN_SPI0;
#else
    /* disable SSC1 */
    *INCA_IP2_PMS_PMS_GEN &= ~INCA_IP2_PMS_PMS_GEN_SPI1;
#endif
    return 0;
}

int ssc_clear_interrupts(void)
{
    *INCA_IP2_ICU_IM2_ISR = 0x3c00;
    return 0;
}

#endif // #if defined(CONFIG_SERIAL_FLASH) || defined(CONFIG_DISPLAY)

