/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

#include <common.h>

#if (CONFIG_COMMANDS & CFG_CMD_NET) && defined(CONFIG_NET_MULTI) \
    && defined(CONFIG_INCA_IP2_SWITCH)
#include <config.h>
#include <malloc.h>
#include <net.h>

#include <asm/addrspace.h>
#include <asm/inca-ip2.h>

#define DMA_WRITE_REG(reg, value) *((volatile u32 *)reg) = (u32)value;
#define DMA_READ_REG(reg, value)    value = (u32)*((volatile u32*)reg)

#define INCAIP2_SW_REG32(reg) *((volatile u32 *)reg)


#define INCAIP2_MDIO_WRITE(phy_id, reg, value) \
                do { \
                        INCAIP2_SW_REG32( SWITCH_MDIO_ACC ) = 0x08000000 | (phy_id << 21) | (reg << 16) | value; \
                        while (!(INCAIP2_SW_REG32( SWITCH_MDIO_ACC ) & 0x80000000)); \
                } while (0)

#define INCAIP2_MDIO_READ(phy_id, reg, value) \
                do { \
                        u32 tmp; \
                        INCAIP2_SW_REG32( SWITCH_MDIO_ACC ) = 0x04000000 | (phy_id << 21) | (reg << 16); \
                        do { \
                                tmp = INCAIP2_SW_REG32( SWITCH_MDIO_ACC ); \
                        } while (!(tmp & 0x80000000)); \
                        value = (u16) tmp; \
                } while (0)

#define NUM_RX_DESC PKTBUFSRX
#define NUM_TX_DESC 20
#define TOUT_LOOP   1000000

/*
#define NUM_RX_DESC 20
#define NUM_TX_DESC 20
*/
#define NUM_SWITCH_DMA_TX_CHANNELS 1
#define NUM_SWITCH_DMA_RX_CHANNELS 1
#define NUM_SWITCH_DMA_CHANNELS (NUM_SWITCH_DMA_TX_CHANNELS + NUM_SWITCH_DMA_RX_CHANNELS)

#define DMA_OWN   1        
#define CPU_OWN   0   

#define DESCR_UNDERRUN 0x00000004
#define SWITCH_DMA_PORT_SELECT   0x00000000
#ifdef __LITTLE_ENDIAN
#define SWITCH_PORT_DEFAULT_CONF 0x00007028
#else /* >> __BIG_ENDIAN */
#define SWITCH_PORT_DEFAULT_CONF 0x00007F28
#endif


#define DMA_RX_CHANNEL_0_SELECT      0x00000000
#define DMA_TX_CHANNEL_1_SELECT      0x00000001
#define DMA_ENABLE_CHANNEL           0x00000001
#define DMA_RESET_ENABLE_CHANNEL     0x00000003
#define DMA_RX_CHANNEL_0_CONTROL     0x00030000
#define DMA_TX_CHANNEL_1_CONTROL     0x00030100
#define DMA_CHANNEL_POLLING_CONTROL  0x80000500 
 
#define DMA_RX_CHANNEL_0_INT_ENABLE   0x0000007E
#define DMA_TX_CHANNEL_1_INT_ENABLE   0x00000000 
#define DMA_IRNEN_RX_CHANNEL_0_ENABLE 0x00000001
 
#define DMA_DESC_OWN_CPU        0x0
#define DMA_DESC_OWN_DMA        0x80000000
#define DMA_DESC_CPT_SET        0x40000000
#define DMA_DESC_SOP_SET        0x20000000
#define DMA_DESC_EOP_SET        0x10000000

#define READ_CMD 0
#define WRITE_CMD 1
#define PHY_STATUS 1
#define PORT0_MDIO_DEV_ID 0x8
#define PORT1_MDIO_DEV_ID 0x9
/* 
#define PARAMETER_CHECK
*/
/* switch related defines */
#define PR_ERR_NONE                     0
#define PR_ERR_TIME_OUT                 74
#define PR_ERR_HASH_FAIL                83

/* defines needed for MAC table access */
#define STATIC              0
#define INVALID             0xF
#define MAX_MACTABLEINDEX   63

#define HASH_CALCULATION_STEPS 4

#define MAC_TBL_VAL     0x2
#define MAC_TABLE_31_0  0
#define MAC_TABLE_55_32 1
#define MAC_MASK        0xFF

#define REQ         1   
#define ACK         0   
#define READPR      1   
#define WRITEPR     0
#define PR_TIME_OUT 10000

/* Critical flag    */
#define     CRITICAL_POS    16
#define     CRITICAL_MASK   0x00010000

/* destination ports */
#define     LAN_PORT_POS   17
#define     LAN_PORT_MASK   0x00020000
#define     PC_PORT_POS    18
#define     PC_PORT_MASK   0x00040000
#define     CPU_PORT_POS   19 
#define     CPU_PORT_MASK   0x00080000
 
/* State    */
#define     MA_STATE_POS    20
#define     MA_STATE_MASK   0xF



#define PR_Reg_Write(address,reg) (*address) = reg
#define PR_Reg_Read(address,reg)   reg = (*address)
    
#define PR_Indirect_Reg_Read(ptr_cmd,ptr_data,ptr_struct,cmd,data) \
        PR_Reg_Write(ptr_cmd,cmd); \
        for (i=0; i<PR_TIME_OUT;i++) \
        { \
            PR_Reg_Read(ptr_cmd,cmd); \
            if(ptr_struct -> Req_ack== ACK) \
            break; \
        } \
        if((ptr_struct -> Req_ack) != ACK) \
            return PR_ERR_TIME_OUT; \
        PR_Reg_Read(ptr_data,data);

#define PR_Indirect_Reg_Read_Sem(ptr_cmd,ptr_data,ptr_struct,cmd,data,ifx_sem) \
        PR_Reg_Write(ptr_cmd,cmd); \
        for (i=0; i<PR_TIME_OUT;i++) \
        { \
            PR_Reg_Read(ptr_cmd,cmd); \
            if(ptr_struct -> Req_ack== ACK) \
            break; \
        } \
        if((ptr_struct -> Req_ack) != ACK) \
            RETURN_ERROR(ifx_sem) \
        PR_Reg_Read(ptr_data,data);
        

#define PR_Indirect_Reg_Write(ptr_cmd,ptr_data,ptr_struct,cmd,data) \
        PR_Reg_Write(ptr_data,data); \
        PR_Reg_Write(ptr_cmd,cmd); \
        for(i=0; i<PR_TIME_OUT;i++) \
        { \
            PR_Reg_Read(ptr_cmd,cmd); \
            if(ptr_struct -> Req_ack == ACK) \
            break; \
        } \
        if((ptr_struct -> Req_ack) != ACK) \
            return PR_ERR_TIME_OUT;



#define PR_Indirect_Reg_Write_Sem(ptr_cmd,ptr_data,ptr_struct,cmd,data,ifx_sem) \
        PR_Reg_Write(ptr_data,data); \
        PR_Reg_Write(ptr_cmd,cmd); \
        for(i=0; i<PR_TIME_OUT;i++) \
        { \
            PR_Reg_Read(ptr_cmd,cmd); \
            if(ptr_struct -> Req_ack == ACK) \
            break; \
        } \
        if((ptr_struct -> Req_ack) != ACK) \
            RETURN_ERROR(ifx_sem)

typedef enum
{
    FALSE,
    TRUE
}
BOOL;


typedef struct 
{
#ifdef __LITTLE_ENDIAN
    union
    {
        struct
        {
            volatile u32 data_length            :16;
            volatile u32 reserve2               :6;
            volatile u32 byte_offset            :2; 
            volatile u32 reserve1               :4;            
            volatile u32 EoP                    :1;
            volatile u32 SoP                    :1;
            volatile u32 C                      :1;
            volatile u32 OWN                    :1;
        }field;
        volatile u32 word;
    }status;
    volatile u32 Data_Pointer;
#else /* __BIG_ENDIAN */
    union
    {
        struct
        {
            volatile u32 OWN                    :1;
            volatile u32 C                      :1;
            volatile u32 SoP                    :1;
            volatile u32 EoP                    :1;
            volatile u32 reserve1               :4;            
            volatile u32 byte_offset            :2; 
            volatile u32 reserve2               :6;
            volatile u32 data_length            :16;
        }field;
        volatile u32 word;
    }status;
    
    volatile u32 Data_Pointer;
#endif
} rx_desc_t; 


typedef struct 
{
#ifdef __LITTLE_ENDIAN
    union
    {
        struct
        {
            volatile u32 data_length     :16;
            volatile u32 reserved        :6;
            volatile u32 byte_offset     :6; 
            volatile u32 EoP             :1;
            volatile u32 SoP             :1;
            volatile u32 C               :1;
            volatile u32 OWN             :1;
        }field;
        volatile u32 word;
    }status;
    volatile u32 Data_Pointer;
#else /* __BIG_ENDIAN */
    union
    {
        struct
        {
            volatile u32 OWN             :1;
            volatile u32 C               :1;
            volatile u32 SoP             :1;
            volatile u32 EoP             :1;
            volatile u32 byte_offset     :6; 
            volatile u32 reserved        :6;
            volatile u32 data_length     :16;
        }field;
        volatile u32 word;
    }status;
    volatile u32 Data_Pointer;
#endif
    
} tx_desc_t;


static rx_desc_t rx_ring[NUM_RX_DESC] __attribute__ ((aligned(64)));  
static tx_desc_t tx_ring[NUM_TX_DESC] __attribute__ ((aligned(64)));  

static int curr_tx_descr = 0;
static int curr_rx_descr = 0;
static int first_time_function_call = 1;

#ifdef __LITTLE_ENDIAN
typedef struct
{
    u32 Req_ack : 1;
    u32 Read_write : 1;
    u32 Table_specific : 4;
    u32 Index : 12;
    u32 Table_sel : 3;
    u32 Rsv : 11;
} cpu_acs_ctrl_reg_t ;
#else /* __BIG_ENDIAN */
typedef struct
{
    u32 Rsv : 11;
    u32 Table_sel : 3;
    u32 Index : 12;
    u32 Table_specific : 4;
    u32 Read_write : 1;
    u32 Req_ack : 1;
} cpu_acs_ctrl_reg_t ;
#endif

typedef u32  cpu_acs_data_reg_t  ;
s32 ifx_get_MAC_table_entry_idx( u8 chipId, u32 macTableIndex, u8* macAddrPtr,
                                 BOOL* lan_portPtr,  BOOL* pc_portPtr, BOOL* cpu_portPtr,
                                 BOOL* criticalFlagPtr, u32* maStatePtr );
s32 ifx_add_MAC_table_entry_idx( u8 chipId, u32 macTableIndex, u8* macAddrPtr, 
                                 BOOL lan_port,  BOOL pc_port, BOOL cpu_port,
                                 BOOL criticalFlag, u32 maState );
typedef unsigned char  crc;

crc  table_crc8[256] = {
    0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15,
    0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d,
    0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65,
    0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d,
    0xe0, 0xe7, 0xee, 0xe9, 0xfc, 0xfb, 0xf2, 0xf5,
    0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
    0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85,
    0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd,
    0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2,
    0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea,
    0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2,
    0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
    0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32,
    0x1f, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0d, 0x0a,
    0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42,
    0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a,
    0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c,
    0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
    0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec,
    0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4,
    0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c,
    0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44,
    0x19, 0x1e, 0x17, 0x10, 0x05, 0x02, 0x0b, 0x0c,
    0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
    0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b,
    0x76, 0x71, 0x78, 0x7f, 0x6a, 0x6d, 0x64, 0x63,
    0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b,
    0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13,
    0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb,
    0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8d, 0x84, 0x83,
    0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb,
    0xe6, 0xe1, 0xe8, 0xef, 0xfa, 0xfd, 0xf4, 0xf3
};

crc crc8(const uint8_t *buf, int count)
{
    crc crc=0;
    int i;

    for(i=0; i<count; i++){
        crc = table_crc8[crc ^ buf[i]];
    }
    return crc;
}

s32 incaip2_write_mdio_access(u8 readwrite, u8 phy_device_id, u8 register_id,
                              u16 *data)
{
    u32 readback = 0x00000000;
    u32 regval = 0x00000000;
    
    if(readwrite==WRITE_CMD)  /* write access */
        regval |= (1<<27);
    else 
        if( readwrite == READ_CMD) /* read access */
            regval |= (1<<26);
        else
            return -1;
        
    regval |= (((u32)phy_device_id & 0x1F) << 21);
    regval |= (((u32)register_id & 0x1F) << 16);    
        
    while(INCAIP2_SW_REG32( SWITCH_MDIO_ACC ) & 0x80000000) ;   

    INCAIP2_SW_REG32( SWITCH_MDIO_ACC ) = regval; 

    if( readwrite == READ_CMD)
    {
        readback = INCAIP2_SW_REG32( SWITCH_MDIO_ACC );
        *data = (u16)(readback & 0x0000FFFF);
    }
    return 0;
}

BOOL mac_addr_equal( u8* macAddr1, u8* macAddr2 )
{
    char i;
    
    for(i=0; i<6;i++)
    {
        if( *(macAddr1+i) != *(macAddr2+i) )
            return 0;
    }
    return 1;
}

/*
 * This function checks whether the entry located at the passed index 
 * is invalid (free) or not.
 * It returns 1 in case of an invalid entry, 0 in case of a valid entry 
 * and -1 in case of an invalid MAC atble access or an invalid parameter.  
 */
s32 isfreeTableIndex( u32 macTableIndex, u8* mac_addr_read )
{

    BOOL lan_port;
    BOOL pc_port;
    BOOL cpu_port;
    BOOL criticalFlag;
    u32 maState;    
    u8 macAddrRead[6]={0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};
    s32 retflag=-1;
    
    if( macTableIndex > MAX_MACTABLEINDEX)
    {
        printf("%s: macTableIndex out of range \n",__FUNCTION__);
        return retflag;
    }

    if ( (retflag = ifx_get_MAC_table_entry_idx( 0, macTableIndex, macAddrRead,     
                                                 &lan_port, & pc_port, &cpu_port,
                                                 &criticalFlag, &maState )) == PR_ERR_NONE )
    {
        
        if( maState == INVALID )
        {
            retflag=1;
        }
        else
        {
            retflag=0;
        }
        memcpy(mac_addr_read, macAddrRead, 6*sizeof(u8));
    }
    else
    {
        printf("%s: ifx_get_MAC_table_entry_idx() failed!\n", __FUNCTION__);
    }
    return retflag;
}

/******************************************************************************
  *
  * FUNCTION:    ifx_get_MAC_hash_idx
  *
  * DESCRIPTION: Get the MAC hash index according to the MAC address.Hash index 
  *              is calculated by dividing the MAC address by the 
  *              polynomial defined in pr_mac.h (for INCAIP2 x^8+x^2+x+1 is used).
  *
  * PARAMETERS:
  *     IN:
  *         macAddrPtr      Pointer to MAC address
  *                         macAddrPtr[0] -> MAC address byte 5
  *                         macAddrPtr[1] -> MAC address byte 4
  *                         macAddrPtr[2] -> MAC address byte 3
  *                         macAddrPtr[3] -> MAC address byte 2
  *                         macAddrPtr[4] -> MAC address byte 1
  *                         macAddrPtr[5] -> MAC address byte 0 
  *                         
  *     OUT:
  *         hashIndexPtr    Pointer to Hash Index (range is 0-MAX_MACTABLEINDEX-1)
  *
  * RETURNS:
  *     PR_ERR_NONE             Success  
  *     PR_ERR_NULL_PTR         Null pointer passed
  *
  ******************************************************************************/
        
s32 ifx_get_MAC_hash_idx( u8* macAddrPtr,  u32* hashIndexPtr )
{   
    u8 my_crc;
    u32 table_index=0xFFFFFFFF;
    s32 free;
    u8 tmp_mac_addr[6];
    u8 mac_addr_read[6]={0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};
    s32 step_count=1;
    u8 bitpos=1;
    BOOL free_index_found = FALSE;
    
    memcpy(&tmp_mac_addr, macAddrPtr, 6*sizeof(u8));

#ifdef PARAMETER_CHECK      
    if(macAddrPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(hashIndexPtr == NULL)
        return PR_ERR_NULL_PTR;
#endif
    do
    {
        /* calculate hash index */
        my_crc = crc8(tmp_mac_addr, 6);
        table_index = my_crc & 0x0000003F;
        /*
         * check for conflict and use manipulated MAC addresses for further hash calculations
         * In case of conflict MAC address bit 0 is inverted and the manipulated  of 
         */
        if( (free=isfreeTableIndex( table_index, mac_addr_read )) == 1)
        {
            free_index_found = TRUE;
        }
        else 
            if( free==0 )
            {
                /*
                 * Requested address is already in table ->
                 * Return index
                 */
                if( mac_addr_equal( macAddrPtr, mac_addr_read ) )
                {
                    free_index_found = TRUE;
                }
                
                /*
                 * Table access conflict ->
                 * Manipulate address and calculate new hash index
                 */
                 if( *(macAddrPtr+5) & bitpos )
                     tmp_mac_addr[5] = (*(macAddrPtr+5) & ~bitpos);
                 else
                     tmp_mac_addr[5] = (*(macAddrPtr+5) | bitpos);
                 bitpos<<=1;
            }
            else
            {
                //printf("%s isfreeTableIndex() returned failure \n",__FUNCTION__);
                return -1;
            }
    } while( !free_index_found && (step_count++ < HASH_CALCULATION_STEPS) );
     
    if(step_count > 3)
    {
   
	    // if there is no free MAC-Index the first used Index is cleaned to be used
	    u8 my_crc;
	    char eth_name_local[NAMESIZE];
	    unsigned char enetaddr_local[6] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
	    u8 mac_addr[6];
	    u32 table_index = {0xFFFFFFFF};

	    memcpy (enetaddr_local, eth_get_dev()->enetaddr, 6);	    
////	    memcpy (eth_name_local, eth_get_dev()->name, NAMESIZE);
	    memcpy(&mac_addr, enetaddr_local, 6*sizeof(u8));
	    my_crc = crc8(&mac_addr, 6);
	    table_index = my_crc & 0x0000003F;
	    ifx_add_MAC_table_entry_idx( 0, table_index, eth_name_local, 0,  0, 1, 0, 0xFF);

    }
    
    if(step_count != 1)
    {
        //printf("Hash calculation conflicts -> needed %d steps !\n", step_count);
    }
    *hashIndexPtr = table_index;
    return PR_ERR_NONE ;
}    

/**
 * Get Mac table index.
 * Get the MAC table index according to the MAC address.
 *
 * \param   macAddrPtr       Pointer to MAC address
 *                             - macAddrPtr[0] -> MAC address byte 5
 *                             - macAddrPtr[1] -> MAC address byte 4
 *                             - macAddrPtr[2] -> MAC address byte 3
 *                             - macAddrPtr[3] -> MAC address byte 2
 *                             - macAddrPtr[4] -> MAC address byte 1
 *                             - macAddrPtr[5] -> MAC address byte 0 
 * \param   mactableIndexPtr Pointer to Table Index (range is 0-63)
 * \return  PR_ERR_NONE      - Success 
 * \return  PR_ERR_NULL_PTR  - macAddrPtr or mactableIndexPtr is null pointer
 * \ingroup Internal
 */
s32 ifx_get_MAC_table_idx( u8* macAddrPtr, u32* mactableIndexPtr )
{   
    u32 hashIndex=0;

#ifdef PARAMETER_CHECK
    if(macAddrPtr == NULL)
    {
        printf("%s failed - invalid macAddrPtr\n", __FUNCTION__);
        return PR_ERR_NULL_PTR;
    }            
    if(mactableIndexPtr == NULL)
    {
        printf("%s failed - invalid mactableIndexPtr\n", __FUNCTION__);
        return PR_ERR_NULL_PTR;
    }                    
#endif
    if(ifx_get_MAC_hash_idx( macAddrPtr, &hashIndex ) != 0)
    {
        printf("%s -> ifx_get_MAC_hash_idx failed !!!\n", __FUNCTION__); 
        return PR_ERR_HASH_FAIL;
    }
    *mactableIndexPtr = hashIndex;
    
    return PR_ERR_NONE ;
    
}   


/**
 * Get MAC table entry index.
 * Get MAC address table entry by MAC address table index.
 *
 * \param   chipId           ChipId
 * \param   macTableIndex    Table Index (range is 0-63)
 * \param   macAddrPtr       Pointer to MAC address
 *                             - macAddrPtr[0] -> MAC address byte 5
 *                             - macAddrPtr[1] -> MAC address byte 4
 *                             - macAddrPtr[2] -> MAC address byte 3
 *                             - macAddrPtr[3] -> MAC address byte 2
 *                             - macAddrPtr[4] -> MAC address byte 1
 *                             - macAddrPtr[5] -> MAC address byte 0 
 * \param   lan_portPtr      TRUE for LAN port
 * \param   pc_portPtr       TRUE for PC port
 * \param   cpu_portPtr      TRUE for CPU port
 * \param   criticalFlagPtr  Pointer to Status of the packet
 * \param   maStatePtr       Pointer to aging/priority data
 * \return  PR_ERR_NONE             - Success 
 * \return  PR_ERR_MAC_TABLE_INDEX  - MAC Table index is invalid 
 * \return  PR_ERR_NULL_PTR         - macAddrPtr, lan_portPtr, pc_portPtr, 
 *                                    cpu_portPtr, criticalFlagPtr, maStatePtr 
 *                                    or base address is null pointer
 * \return  PR_ERR_TIME_OUT         - Time out
 * \return  PR_ERR_CHIP_ID          - Invalid Chip Id
 * \ingroup Internal
 */
s32 ifx_get_MAC_table_entry_idx( u8 chipId, u32 macTableIndex, u8* macAddrPtr,
                                        BOOL* lan_portPtr,  BOOL* pc_portPtr, BOOL* cpu_portPtr,
                                        BOOL* criticalFlagPtr, u32* maStatePtr )
{
    u32 hashIndex =0;
    u32 Temp = 0;
    s32 i =0;
    /* pointer to CPU access control register */
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;  
    /* pointer to CPU access data register */
    u32* ptr_cpu_acs_data;              
    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK
    if(macTableIndex > MAX_MACTABLEINDEX)
        return PR_ERR_MAC_TABLE_INDEX;
    if((macAddrPtr == NULL) ||
       (lan_portPtr == NULL) ||
       (pc_portPtr == NULL) ||
       (cpu_portPtr == NULL) ||
       (criticalFlagPtr == NULL) ||
       (maStatePtr == NULL) )
        return PR_ERR_NULL_PTR;
#endif

    ptr_cpu_acs_data = (u32*) CPU_ACS_DATA_REG;

    ptr_reg1 = (u32*) CPU_ACS_CTRL_REG;

    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    /*
     * read lower 32 bits of MAC table entry from MAC table
     */
    hashIndex = macTableIndex; /*  >> HASHIND_POS;  */    
    ptr_cpu_acs_ctrl->Index          = hashIndex; 
    ptr_cpu_acs_ctrl->Table_specific = MAC_TABLE_31_0;    
    ptr_cpu_acs_ctrl->Table_sel      = MAC_TBL_VAL;
    ptr_cpu_acs_ctrl->Read_write     = READPR;  /* read */    
    ptr_cpu_acs_ctrl->Req_ack        = REQ;    

    /* First read operation, Data reg -> MAC3 MAC2 MAC1 MAC0 */ 
    PR_Indirect_Reg_Read( ptr_reg1, ptr_cpu_acs_data,
                                  ptr_cpu_acs_ctrl, temp1, Temp );
    for(i = 5; i >= 2; i--)
        *(macAddrPtr+i) = (Temp >> (8 * (5-i))) & MAC_MASK;
        
    /*
     * read upper 32 bits of MAC table entry from MAC table
     * These are: MAC address bits 32...55, critical bit, LAN, PC and 
     * CPU port destination bits, aging state 
     */
    Temp=0;
    
    ptr_cpu_acs_ctrl->Index          = hashIndex; 
    ptr_cpu_acs_ctrl->Table_sel      = MAC_TBL_VAL;
    ptr_cpu_acs_ctrl->Read_write     = READPR;  /* read */    
    ptr_cpu_acs_ctrl->Table_specific = MAC_TABLE_55_32;    
    ptr_cpu_acs_ctrl->Req_ack        = REQ;    

    PR_Indirect_Reg_Read( ptr_reg1, ptr_cpu_acs_data,
                                  ptr_cpu_acs_ctrl, temp1, Temp );
    /* 
     * get destination port, critical bit and aginf state information
     * from read data word 
     */                         
    *lan_portPtr = (Temp & LAN_PORT_MASK) !=0 ? TRUE : FALSE;       
    *pc_portPtr = (Temp & PC_PORT_MASK) !=0 ? TRUE : FALSE;     
    *cpu_portPtr = (Temp & CPU_PORT_MASK) !=0 ? TRUE : FALSE;     
    *criticalFlagPtr = (Temp & CRITICAL_MASK) !=0 ? TRUE : FALSE;           
    *maStatePtr = (Temp >> MA_STATE_POS) &  MA_STATE_MASK;   
    
    /* MAC address byte 4 & 5 */
    *(macAddrPtr+1) = Temp & MAC_MASK;  
    *macAddrPtr = (Temp >> 8) & MAC_MASK;

    return PR_ERR_NONE ;
    
}

/**
 * Add MAC table entry.
 * Add MAC address table entry by MAC address table index.
 *
 * \param   chipId           ChipId
 * \param   macTableIndex    Table Index (range is 0-63)
 * \param   macAddrPtr       Pointer to MAC address
 *                             - macAddrPtr[0] -> MAC address byte 5
 *                             - macAddrPtr[1] -> MAC address byte 4
 *                             - macAddrPtr[2] -> MAC address byte 3
 *                             - macAddrPtr[3] -> MAC address byte 2
 *                             - macAddrPtr[4] -> MAC address byte 1
 *                             - macAddrPtr[5] -> MAC address byte 0 
 * \param   lan_port         TRUE for LAN port
 * \param   pc_port          TRUE for PC port
 * \param   cpu_port         TRUE for CPU port
 * \param   criticalFlag     Status of the packet
 * \param   maState          aging/priority data
 * \return  PR_ERR_NONE             - Success 
 * \return  PR_ERR_AGING_STATE      - Invalid aging state
 * \return  PR_ERR_FILTERID         - If the Filter Id is invalid 
 * \return  PR_ERR_MAC_TABLE_INDEX  - MAC Table index is invalid 
 * \return  PR_ERR_CRITICAL_FLAG    - Critical flag invalid   
 * \return  PR_ERR_NULL_PTR         - macAddrPtr or base address is null pointer
 * \return  PR_ERR_TIME_OUT         - Time out
 * \return  PR_ERR_CHIP_ID          - Invalid Chip Id
 * \ingroup Internal
 */
s32 ifx_add_MAC_table_entry_idx( u8 chipId, u32 macTableIndex, u8* macAddrPtr, 
                                 BOOL lan_port,  BOOL pc_port, BOOL cpu_port,
                                 BOOL criticalFlag, u32 maState )
{   
    u32 hashIndex =0;
    u32 Temp = 0;
    s32 i=0;
    
    /* pointer to CPU access control register */
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    /* pointer to CPU access data register */
    u32* ptr_cpu_acs_data;

    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK  
    if(macTableIndex > MAX_MACTABLEINDEX)
    {
        printf("%s failed - invalid macTableIndex %d\n", __FUNCTION__, macTableIndex);
        return PR_ERR_MAC_TABLE_INDEX;
    }
    if(maState > MAX_AGING_STATE)
    {
        printf("%s failed - invalid aging state %d\n", __FUNCTION__, maState);
        return PR_ERR_AGING_STATE;
    }        
    if((criticalFlag != TRUE) && (criticalFlag != FALSE))
    {
        printf("%s failed - invalid criticalFlag %d\n", __FUNCTION__, criticalFlag);
        return PR_ERR_CRITICAL_FLAG;
    }        
    if(macAddrPtr == NULL)
    {
        printf("%s failed - invalid macAddrPtr\n", __FUNCTION__);
        return PR_ERR_NULL_PTR;
    }            
#endif

    ptr_cpu_acs_data = (u32*) CPU_ACS_DATA_REG;
    ptr_reg1 = (u32*) CPU_ACS_CTRL_REG;
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    /*
     * write lower 32 bits of MAC table entry to MAC table
     */
    Temp=0;
    for(i = 2; i <= 5; i++)
        Temp |= (*(macAddrPtr+i) << (8 * (5-i)));
    *ptr_cpu_acs_data = Temp;
    /* Extract hash index from Table Index */
    /* HASHIND_POS = 2 */
    hashIndex = macTableIndex;  /*  >> HASHIND_POS;  */    
    ptr_cpu_acs_ctrl->Index          = hashIndex;
    ptr_cpu_acs_ctrl->Table_specific = MAC_TABLE_31_0;    
    ptr_cpu_acs_ctrl->Table_sel      = MAC_TBL_VAL;
    ptr_cpu_acs_ctrl->Read_write     = WRITEPR;  /* write */    
    ptr_cpu_acs_ctrl->Req_ack        = REQ;    

    PR_Indirect_Reg_Write( ptr_reg1,         /* CPU_ACS_CTRL_REG */
                               ptr_cpu_acs_data,
                               ptr_cpu_acs_ctrl, 
                               temp1,
                               Temp );
    /*
     * write upper 32 bits of MAC table entry to MAC table
     * These are: MAC address bits 32...55, critical bit, LAN, PC and 
     * CPU port destination bits, aging state 
     */
    Temp=0;
    for(i = 0; i <= 1; i++)  /* MAC address byte 5 & 4 */
        Temp |= (*(macAddrPtr+i) << (8 * (1-i)));
    
    /* set critical bit acc. to passed parameter */
    Temp |= criticalFlag ? (1<< CRITICAL_POS) : 0;        
    /* set port destination bits acc. to passed parameters */
    Temp |= lan_port ? (1<< LAN_PORT_POS) : 0;     
    Temp |= pc_port ? (1<< PC_PORT_POS) : 0;       
    Temp |= cpu_port ? (1<< CPU_PORT_POS) : 0;         
    /* set state value */
    Temp |= (maState << MA_STATE_POS);    
    *ptr_cpu_acs_data = Temp;  
    
    hashIndex = macTableIndex;  /*  >> HASHIND_POS;  */    
    ptr_cpu_acs_ctrl->Index          = hashIndex;
    ptr_cpu_acs_ctrl->Table_sel      = MAC_TBL_VAL;
    ptr_cpu_acs_ctrl->Table_specific = MAC_TABLE_55_32;  
    ptr_cpu_acs_ctrl->Read_write     = WRITEPR; 
    ptr_cpu_acs_ctrl->Req_ack        = REQ;  

    PR_Indirect_Reg_Write( ptr_reg1, ptr_cpu_acs_data,
                               ptr_cpu_acs_ctrl, temp1, 
                               Temp );
           
    return PR_ERR_NONE ;
}

s32 incaip2_mac_table_add_entry( u8* macAddrPtr, BOOL lan_port, BOOL pc_port, BOOL cpu_port,
                                 BOOL criticalFlag, u32 maState )
{
    u32 macTableIndex;
    s32 retflag = -1;
    
    BOOL lan_portRead;  
    BOOL pc_portRead; 
    BOOL cpu_portRead;
    BOOL criticalFlagRead; 
    u32  maStateRead;
    u8 macAddr_read[6]={0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};    

    /* 
     * get table index and insert new entry at calculated position 
     */
    if( (retflag = ifx_get_MAC_table_idx( macAddrPtr, &macTableIndex ))
        == PR_ERR_NONE )
    {
        retflag = ifx_add_MAC_table_entry_idx( 0, /* chipid = 0 */
                                               macTableIndex,
                                               macAddrPtr,
                                               lan_port,
                                               pc_port, 
                                               cpu_port, 
                                               criticalFlag,
                                               maState );   
                                       
        if(retflag== PR_ERR_NONE)  
        {                                  
            if( (retflag = ifx_get_MAC_table_entry_idx( 0, 
                                                       macTableIndex, 
                                                       macAddr_read,
                                                       &lan_portRead,  
                                                       &pc_portRead, 
                                                       &cpu_portRead,
                                                       &criticalFlagRead, 
                                                       &maStateRead )) == PR_ERR_NONE)
            {
                if( !mac_addr_equal( macAddrPtr, macAddr_read ) )
                {
                    printf("%s - read address different from written\n", __FUNCTION__);
                    printf("written %02X:%02X:%02X:%02X:%02X:%02X at index %d\n", 
                           *macAddrPtr, *(macAddrPtr+1), *(macAddrPtr+2),*(macAddrPtr+3),
                           *(macAddrPtr+4), *(macAddrPtr+5), macTableIndex);
                    printf("read %02X:%02X:%02X:%02X:%02X:%02X at index %d\n", 
                           *macAddr_read, *(macAddr_read+1), *(macAddr_read+2),*(macAddr_read+3),
                           *(macAddr_read+4), *(macAddr_read+5), macTableIndex);                           
                           
                    retflag = -1;
                }   
                if(lan_port != lan_portRead) 
                {
                    printf("%s - read lan_port (%d) different from written (%d)\n",
                           __FUNCTION__, lan_portRead, lan_port);  
                    retflag = -1;
                }   
                if(pc_port != pc_portRead) 
                {
                    printf("%s - read pc_port (%d) different from written (%d)\n", 
                           __FUNCTION__,pc_portRead, pc_port);
                    retflag = -1;
                }                  
                if(cpu_port != cpu_portRead) 
                {
                    printf("%s - read cpu_port (%d) different from written  (%d)\n", 
                           __FUNCTION__, cpu_portRead, cpu_port);
                    retflag = -1;
                }                   
                if(criticalFlag != criticalFlagRead) 
                {
                    printf("%s - read criticalFlag (%d) different from written (%d)\n", 
                           __FUNCTION__,criticalFlagRead,criticalFlag);
                    retflag = -1;
                }                    
                if(maState != maStateRead) 
                {
                    printf("%s - read maState (0x%X) different from written (0x%X)\n",
                           __FUNCTION__, maStateRead, maState);
                    retflag = -1;
                } 
                
                if(retflag == -1)
                {
                    printf("%s read back entry read at pos. %d differs from written\n", __FUNCTION__, macTableIndex);
                    return retflag;
                }
                else
                {
                    retflag = PR_ERR_NONE;
                }
            }
            else
            {
                printf("%s read back mac table entry at index %d failed\n", __FUNCTION__, macTableIndex);
            }
        }
                                              
    }
    else
        printf("%s -> ifx_get_MAC_table_idx failed\n", __FUNCTION__);

    return retflag;
}

void INCAIP2_Switch_Print_MAC_TableHeader(void)
{

    printf("\nINCAIP2 MAC Table\n");
    printf("------------------------------------------------------------------\n");
    printf("| index |   MAC Address      || LAN PC  CPU crit  age           \n");
    printf("------------------------------------------------------------------\n");

}

void INCAIP2_Switch_Print_MAC_TableEntry( u32 macTableIndex, u8 *macAddr, 
                                          BOOL lan_port,  BOOL pc_port, BOOL cpu_port,
                                          BOOL criticalFlag, u32 maState )
{
      printf("[%3d] |  %02x:%02x:%02x:%02x:%02x:%02x    || %1c | %1c | %1c | %1c | %3d\n",
              macTableIndex,
              macAddr[0], macAddr[1], macAddr[2],
              macAddr[3], macAddr[4], macAddr[5],
              (lan_port)?'+':' ',  (pc_port)?'+':' ', (cpu_port)?'+':' ', 
              (criticalFlag)?'1':'0', maState );
}

void INCAIP2_Switch_Print_MAC_TableTrailer(int count)
{
   /*
          "        INCA-IP MAC table entries"
          "INDEX |     MAC Address    || Cntrl -> |M|V|L|P|CR| age "
          "----------------------------------------------------------------"
           [182] |  00:50:bf:49:eb:83 ||  0xe2 -> | | | | |  | 111 aged out
           [212] |  00:a0:24:b8:b8:47 ||  0x14 -> | | | | |  | 000 static
           [182] |  00:50:bf:49:eb:83 ||  0xe2 -> | | | | |  | 111 dynamic
   */

   printf("----------------------------------------------------------------\n") ;
   printf("  ==>  MAC TABLE: %d entries found\n", count) ;
   printf("----------------------------------------------------------------\n") ;
}

void INCAIP2_Switch_Print_MAC_Table(void)
{
    u8 chipId = 0;
    u32 macTableIndex;
    u8 macAddr[6];
    BOOL lan_port;
    BOOL pc_port;    
    BOOL cpu_port;    
    BOOL criticalFlag;
    u32 maState;
    s32 count = 0;

    INCAIP2_Switch_Print_MAC_TableHeader();
    for ( macTableIndex = 0; macTableIndex <= MAX_MACTABLEINDEX; macTableIndex++ )
    {
        if( ifx_get_MAC_table_entry_idx( chipId, macTableIndex, macAddr,
                                         &lan_port,  &pc_port, &cpu_port,
                                         &criticalFlag, &maState ) != 0 )
        {
            printf("ifx_get_MAC_table_entry_idx failed for index %d\n", macTableIndex);
            return;
        }

        if(maState!=0xF)
        {
            count++;

            INCAIP2_Switch_Print_MAC_TableEntry( macTableIndex, macAddr, 
                                                 lan_port,  pc_port, cpu_port,
                                                 criticalFlag, maState );
        }

    }
    INCAIP2_Switch_Print_MAC_TableTrailer(count);
    return;
}

static u32 incaip2_dma_read_channel_control_reg(u32 channel)
{
    if(channel>19)
    {
        printf("%s - Invalid Channel passed\n", __FUNCTION__);
        return 0;
    }
    
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)  = channel;
    return ((u32)*((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CCTRL));
}

static u32 incaip2_dma_read_channel_descr_base(u32 channel)
{
    if(channel>19)
    {
        printf("%s - Invalid Channel passed\n", __FUNCTION__);
        return 0;
    }
    
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)  = channel;
    return ((u32)*((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CDBA));
}

static void incaip2_dma_disable_channel_interrupts(void)
{
    /*
     * Disable Interrupts for DMA channels 0 and 1 
     */
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)  = DMA_RX_CHANNEL_0_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CIE) = 0x00000000;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)  = DMA_TX_CHANNEL_1_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CIE) = 0x00000000; 

    *((volatile u32*)INTERRUPT_NODE_REGISTERS_DMA_IRNEN) = 0x00000000;;    
}


/**
 * Enable DMA channel Interrupts for rx cannnel 0
 * Enables the DMA rx channel 0 interrupt at the interrupt node enable register
 * 
 * \param  None
 * \return Nothing
 */
static void incaip2_dma_enable_channel_interrupts(void)
{
    /*
     * Enable Interrupts for DMA channels 0 and 1 
     */
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)  = DMA_RX_CHANNEL_0_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CIE) = DMA_RX_CHANNEL_0_INT_ENABLE;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)  = DMA_TX_CHANNEL_1_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CIE) = DMA_TX_CHANNEL_1_INT_ENABLE; 
    
    *((volatile u32*)INTERRUPT_NODE_REGISTERS_DMA_IRNEN) = DMA_IRNEN_RX_CHANNEL_0_ENABLE;
}


static void incaip2_dma_disable_channels( void )
{
    /*
     * disable all dma channels used by switch 
     * by clearing channel's enable bit
     */ 
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_RX_CHANNEL_0_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CCTRL) &= ~(DMA_ENABLE_CHANNEL);
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_TX_CHANNEL_1_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CCTRL) &= ~(DMA_ENABLE_CHANNEL); 
}


static void incaip2_dma_enable_channels( void )
{
    /*
     * disable all dma channels used by switch 
     * by clearing channel's enable bit
     */ 
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_RX_CHANNEL_0_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CCTRL) |= DMA_ENABLE_CHANNEL;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_TX_CHANNEL_1_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CCTRL) |= DMA_ENABLE_CHANNEL; 
}

static void incaip2_dma_reset_enable_tx_channel( void )
{
    /*
     * disable all dma channels used by switch 
     * by clearing channel's enable bit
     */ 
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_TX_CHANNEL_1_SELECT;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CCTRL) |= DMA_RESET_ENABLE_CHANNEL; 
}

/**
 * Initialize DMA Registers for for 2 channel switch DMA operation
 * This function configures the DMA channel and port specific registers 
 * to setup communication with the switch module via high priority rx channel 0 
 * and tx channel 1.
 * 
 * \param  None
 * \return Nothing
 */
static void incaip2_dma_init(void)
{

    *((volatile u32*)INCA_IP2_PMS_PMS_GEN) |= INCA_IP2_PMS_PMS_GEN_DMA;       
    /*
     * Disable DMA channels and related interrupts
     */
    incaip2_dma_disable_channels();
    incaip2_dma_disable_channel_interrupts();
             
    /*
     * configure switch port
     */
    *((volatile u32*)PORT_RELATED_REGISTERS_DMA_PS)    = SWITCH_DMA_PORT_SELECT;    
    *((volatile u32*)PORT_RELATED_REGISTERS_DMA_PCTRL) = SWITCH_PORT_DEFAULT_CONF;     
    
    /*
     * Set channel descriptor base addresses and length for switch channels
     */
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)   = DMA_RX_CHANNEL_0_SELECT;    
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CDBA) = (u32) PHYSADDR(&rx_ring);
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CDLEN)= (u32) NUM_RX_DESC;

    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)   = DMA_TX_CHANNEL_1_SELECT;    
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CDBA) = (u32) PHYSADDR(&tx_ring);
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CDLEN)= (u32) NUM_TX_DESC;
    
    /*
     * Set Channel Parameters but leave channels disabled
     */
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_RX_CHANNEL_0_SELECT;         
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CCTRL) = DMA_RX_CHANNEL_0_CONTROL;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_TX_CHANNEL_1_SELECT;         
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CCTRL) = DMA_TX_CHANNEL_1_CONTROL;
 
    /*
     * Enable polling mode for both channels
     */    
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_RX_CHANNEL_0_SELECT;      
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CPOLL) = DMA_CHANNEL_POLLING_CONTROL;
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)    = DMA_TX_CHANNEL_1_SELECT;       
    *((volatile u32*)CHANNEL_RELATED_REGISTERS_DMA_CPOLL) = DMA_CHANNEL_POLLING_CONTROL; 
    
}
/************ NEW ******************************************************************************/
#define MDIO_INT_EXT_MASK       0x00000020
#define MDIO_INT_EXT_POS        5
#define SWITCH_GMAC_LG          0x00000010

#undef IFX_DEBUG_MDIO_STATE_MACHINE
#ifdef IFX_DEBUG_MDIO_STATE_MACHINE
static char* ifx_switch_phy_speed_names[]={ "10MBit/s", "100MBit/s","1000MBits/s"};
static char* ifx_switch_phy_duplex_names[]={ "half duplex", "full duplex"};
static char* ifx_switch_phy_pause_names[]={ "no pause", "with pause"};
#endif

/**
 * Update GMAC register settings.
 * This funtion update the values in GMAC according to the values in the phy 
 * registers.
 *
 * \param   phy_id     MDIO id of phy
 * \param   gmac_value Current value of GMAC register
 */
u32 ifx_switch_update_gmac_autoneg(int phy_id,u32 gmac_value)
{
     int chip_id = 0;
     int port_id = 0;
     u8 phy_duplex=0;
     u8 phy_pause=0;
     u8 phy_speed=0;
     BOOL mode;
     u16 phy_reg4=0;	// Autonegotiation Copper Advertisement
     u16 phy_reg5=0;	// Autonegotiation Copper Link Partner Ability - Base Page
     u16 phy_reg9=0;	// 1000BASE-T Control
     u16 phy_reg10=0;	// 1000BASE-T Status
     u32 reg=0;
     
//     if (phy_id == 1) 
//	     phy_id = 0;
//     else 
//	     phy_id = 1;
     
     
     INCAIP2_MDIO_READ(phy_id, 4, phy_reg4);
     INCAIP2_MDIO_READ(phy_id, 5, phy_reg5);
     
     mode = ((INCAIP2_SW_REG32( MDIO_MOD_SEL ) & MDIO_INT_EXT_MASK) >> MDIO_INT_EXT_POS);    
//     ifx_get_MDIO_phy(chip_id,port_id,&mode);
     if (mode==1) {
          /* external phy used => 1000 MBit */
          
          /* Read register #9 for GBit advertisement settings */
          INCAIP2_MDIO_READ(phy_id, 9, phy_reg9);
          
          /* Read register #10 for GBit partner settings */
          INCAIP2_MDIO_READ(phy_id, 10, phy_reg10);
     }
     
#ifdef IFX_DEBUG_MDIO_STATE_MACHINE
     printf("TRACE %s %s %d mode=%d\n",__FILE__,__FUNCTION__,__LINE__,mode);
     printf("TRACE %s %s %d PHY Registers: #4=0x%08x #5=0x%08x #9=0x%08x #10=0x%08x\n",__FILE__,__FUNCTION__,__LINE__,phy_reg4,phy_reg5,phy_reg9,phy_reg10);
#endif

     /* Check if advertise and partner agree on pause */
     phy_pause = ((phy_reg4 & phy_reg5) & 0x0400) >> 10;
     
     /* Find the best mode both partners support */
     /* Priority: 1GB-FD, 1GB-HD, 100MB-FD, 100MB-HD, 10MB-FD, 10MB-HD */
     phy_speed = ((((phy_reg9<<2) & phy_reg10) & 0x0c00) >> 6) |
                 (((phy_reg4 & phy_reg5) & 0x01e0) >> 5);
     
     if (phy_speed >= 0x0020) {
          phy_speed = 2;
          phy_duplex = 1;
     }
     else if (phy_speed >= 0x0010) {
          phy_speed = 2;
          phy_duplex = 0;
     }
     else if (phy_speed >= 0x0008) {
          phy_speed = 1;
          phy_duplex = 1;
     }
     else if (phy_speed >= 0x0004) {
          phy_speed = 1;
          phy_duplex = 0;
     }
     else if (phy_speed >= 0x0002) {
          phy_speed = 0;
          phy_duplex = 1;
     }
     else if (phy_speed >= 0x0001) {
          phy_speed = 0;
          phy_duplex = 0;
     }
     if (phy_duplex == 0) phy_pause = 1;
     reg=(gmac_value & ~0x0000006e) | (phy_pause << 6) |
                      (phy_pause << 5) | (phy_duplex << 3) | (phy_speed << 1);
#ifdef IFX_DEBUG_MDIO_STATE_MACHINE
//printk("TRACE %s %s %d speed=%d duplex=%d pause=%d reg=0x%08x\n",__FILE__,__FUNCTION__,__LINE__,phy_speed,phy_duplex,phy_pause,reg);
     if ( ((phy_speed & ~0x3)==0)  && ((phy_duplex & ~0x1)==0) && ((phy_pause & ~0x1)==0) ) { 
          printf("TRACE %s %s %d GMAC set to %s %s %s\n",__FILE__,__FUNCTION__,__LINE__,ifx_switch_phy_speed_names[phy_speed],ifx_switch_phy_duplex_names[phy_duplex],ifx_switch_phy_pause_names[phy_pause]);
     }
#endif

     return reg;
     
}

/**
 * Autonegotiation check.
 * This funtion checks for link changes. If a link change has occured it will 
 * update certain switch registers.
 */
void ifx_switch_check_phy_status(void)
{
     static int link_loss_detected_port0=1;
     static int link_loss_detected_port1=1;
     int chip_id = 0;
     int port_id = 0;
	u32 reg=0;
     u16 phy_reg=0;
     u16 phy_link=0;
     u8 phy_id=0;
     u8 phy_duplex;
     u8 phy_pause;
     u8 phy_speed;
     
     static int counter=0;
          
     counter++;
#ifdef IFX_DEBUG_MDIO_STATE_MACHINE
     if ((counter % 10000) == 0) printf("TRACE %s %s called %d times\n",__FILE__,__FUNCTION__,counter);
#endif

     /* we need to check both ports */
     
     /* First we check LAN port */
     port_id=0;
     phy_id=PHY_LAN;
     INCAIP2_MDIO_READ(phy_id, 0, phy_reg);
	
     /* Is autoneg enabled? */
     if (phy_reg & 0x1000) {
          reg=INCAIP2_SW_REG32(SWITCH_P0_GMAC_REG);
          if ((reg & SWITCH_GMAC_LG)==0) {
               link_loss_detected_port0 = 1;
          } 
          else {
               if (link_loss_detected_port0) {
#ifdef IFX_DEBUG_MDIO_STATE_MACHINE
printf("TRACE %s %s %d link loss detected port %d\n",__FILE__,__FUNCTION__,__LINE__,port_id);
#endif
                    /* transfer phy registers settings to switch */
                    reg=ifx_switch_update_gmac_autoneg(phy_id,reg);
                    INCAIP2_SW_REG32(SWITCH_P0_GMAC_REG)=reg;
                    link_loss_detected_port0 = 0;
               }
          
          }
     }
     /* We are in fixed state */
     else {
          /* Check link status in phy register #1 */
          INCAIP2_MDIO_READ(phy_id, 1, phy_link);
          if (phy_link & 0x0004) {
               /* duplex and speed is stored in phy register #0 */
               phy_speed = ( ((phy_reg & 0x0040) >> 5) | ((phy_reg & 0x2000) >> 13));
               phy_duplex = ((phy_reg & 0x0100) >> 8);
               if (phy_duplex == 0) {
                    phy_pause = 1;
               }
               else {
                    /* Read phy register #4 for pause values */
                    INCAIP2_MDIO_READ(phy_id, 4, phy_reg);
                    phy_pause = ((phy_reg & 0x0400) >> 10);
               }
               
               /* Set bits in GMAC register (Link status is ok) */
               reg=INCAIP2_SW_REG32(SWITCH_P0_GMAC_REG);
               reg=(reg & ~(0x0000007e)) | (phy_pause<<6) | (phy_pause<<5) |
                    (phy_duplex << 3) | (phy_speed << 1) | 0x00000010;
               INCAIP2_SW_REG32(SWITCH_P0_GMAC_REG)=reg;
          }
          else {
               /* Set link status in GMAC register to not ok */
               reg=INCAIP2_SW_REG32(SWITCH_P0_GMAC_REG);
               reg=(reg & ~(0x00000010));
               INCAIP2_SW_REG32(SWITCH_P0_GMAC_REG)=reg;
          }               
     }     
     
    
     /* Now we check PC port */
     port_id = 1;
     phy_id=PHY_PC;
     /* Read phy register #0 */
     INCAIP2_MDIO_READ(phy_id, 0, phy_reg);

     /* Is autoneg enabled? */
     if (phy_reg & 0x1000) {
          reg=INCAIP2_SW_REG32(SWITCH_P1_GMAC_REG);
          if ((reg & SWITCH_GMAC_LG)==0) {
               link_loss_detected_port1 = 1;
          }
          else {
              if (link_loss_detected_port1) {
#ifdef IFX_DEBUG_MDIO_STATE_MACHINE
printf("TRACE %s %s %d link loss detected port %d\n",__FILE__,__FUNCTION__,__LINE__,port_id);
#endif
                    /* transfer phy registers settings to switch */
                    reg=ifx_switch_update_gmac_autoneg(phy_id,reg);
                    INCAIP2_SW_REG32(SWITCH_P1_GMAC_REG)=reg;
                    link_loss_detected_port1 = 0;
               }
          
          }
     }
     /* We are in fixed state */
     else {
          /* Check link status in phy register #1 */
          INCAIP2_MDIO_READ(phy_id, 1, phy_link);
          if (phy_link & 0x0004) {
               /* duplex and speed is stored in phy register #0 */
               phy_speed = ( ((phy_reg & 0x0040) >> 5) | ((phy_reg & 0x2000) >> 13));
               phy_duplex = ((phy_reg & 0x0100) >> 8);
               if (phy_duplex == 0) {
                    phy_pause = 1;
               }
               else {
                    /* Read phy register #4 for pause values */
                    INCAIP2_MDIO_READ(phy_id, 4, phy_reg);
                    phy_pause = ((phy_reg & 0x0400) >> 10);
               }
               
               /* Set bits in GMAC register (Link status is ok) */
               reg=INCAIP2_SW_REG32(SWITCH_P1_GMAC_REG);
               reg=(reg & ~(0x0000007e)) | (phy_pause<<6) | (phy_pause<<5) |
                    (phy_duplex << 3) | (phy_speed << 1) | 0x00000010;
               INCAIP2_SW_REG32(SWITCH_P1_GMAC_REG)=reg;
          }
          else {
               /* Set link status in GMAC register to not ok */
               reg=INCAIP2_SW_REG32(SWITCH_P1_GMAC_REG);
               reg=(reg & ~(0x00000010));
               INCAIP2_SW_REG32(SWITCH_P1_GMAC_REG)=reg;
          }               
         
     }
     
}

static void incaip2_print_rx_descriptors(void )
{
    s32 i;
    rx_desc_t* rx_desc_p;
    
    printf(" i  &rx_ring[i]   Status  Data Ptr\n");    
    printf("--------------------------------------\n");

    for (i = 0; i < NUM_RX_DESC; i++) 
    {
        rx_desc_p=  (rx_desc_t*) KSEG1ADDR(&rx_ring[i]);    
        printf("%d  0x%08X   0x%08X   0x%08X", 
                 i, (u32)&rx_ring[i], rx_desc_p->status, rx_desc_p->Data_Pointer );
        if(curr_rx_descr==i)
            printf(" <- CURRENT\n");
        else
            printf("\n");
    }    
}


static void incaip2_print_tx_descriptors(void )
{
    s32 i;
    tx_desc_t* tx_desc_p;   
    
    printf(" i &tx_ring[i]  Status  Data Ptr\n");    
    printf("--------------------------------------\n");            
    for (i = 0; i < NUM_TX_DESC; i++) 
    {
        tx_desc_p=  (tx_desc_t*) KSEG1ADDR(&tx_ring[i]);            
        printf("%d  0x%08X   0x%08X   0x%08X", 
                 i, (u32)&tx_ring[i], tx_desc_p->status, tx_desc_p->Data_Pointer );
        if(curr_tx_descr==i)
            printf(" <- CURRENT\n");    
        else
            printf("\n");        
    }

}

static void incaip2_print_descriptors(void )
{
    incaip2_print_rx_descriptors();
    incaip2_print_tx_descriptors();
}

static int incaip2_switch_init(struct eth_device *dev, bd_t * bis)
{
    rx_desc_t* rx_desc_p;
    tx_desc_t* tx_desc_p;   
    s32 i;
    u8 own_mac_addr[6];
    BOOL lan_port, pc_port, cpu_port;    
    BOOL criticalFlag;  
    u32 maState;

    /*
     * Add own MAC address which is passed in eth device structure
     */
    memcpy(own_mac_addr, dev->enetaddr, 6);
    /* add own MAC address, overwrite existing */
    if( incaip2_mac_table_add_entry( own_mac_addr,
                                     lan_port = FALSE, 
                                     pc_port = FALSE, 
                                     cpu_port = TRUE,
                                     criticalFlag = TRUE,
                                     maState = STATIC ) != 0)
    {
        printf("%s - Adding MAC table entry failed\n", __FUNCTION__);
    }

    /*
     * Initialize rx and tx descriptor chains for rx channel 0 and tx channel 1 
     */
    if( first_time_function_call)
    {  
        for (i = 0; i < NUM_RX_DESC; i++) 
        {
            rx_desc_p=  (rx_desc_t*) KSEG1ADDR(&rx_ring[i]);
            memset(rx_desc_p, 0, sizeof(tx_ring[i]));         
            rx_desc_p->status.word = 0;
            rx_desc_p->status.field.data_length=PKTSIZE_ALIGN;
            rx_desc_p->Data_Pointer = (u32)KSEG1ADDR(NetRxPackets[i]);
            rx_desc_p->status.field.byte_offset=0;
            rx_desc_p->status.field.OWN=DMA_OWN;
        } 
        first_time_function_call = 0;
    }
 
    for (i = 0; i < NUM_TX_DESC; i++) 
    {
        tx_desc_p = (tx_desc_t*) KSEG1ADDR(&tx_ring[i]);
        memset(tx_desc_p, 0, sizeof(tx_ring[i]));

        tx_desc_p->status.word = 0;
        tx_desc_p->status.field.data_length=PKTSIZE_ALIGN;
        tx_desc_p->status.field.OWN=CPU_OWN;  
        tx_desc_p->status.field.C = 1;       
    }
    /* set initial descriptor position for tx descriptor */
    curr_tx_descr = 0;

    /* forward unknown multicast/unicast only to LAN ports */
    /* forward unknown broadcast to all ports (incl. CPU)  */
    INCAIP2_SW_REG32( UNKNOWN_DEST_REG ) = 0x000000DF;
    /* Enable Switch Ports in forwarding mode and enable DMA channels & interrupts */
    INCAIP2_SW_REG32( ST_PT_REG )=0x0000003F;
    incaip2_dma_enable_channel_interrupts();    
    incaip2_dma_reset_enable_tx_channel();

   ifx_switch_check_phy_status();
    
    return 1;
}



static int incaip2_switch_send(struct eth_device *dev, volatile void *packet, 
                               int length)
{
   int            res  = -1;
    tx_desc_t      *tx_desc = KSEG1ADDR(&tx_ring[curr_tx_descr]);
    int i;
    s32 byte_offset=0;    
    u32 tx_chan_interrupts;   
    static unsigned int send_counter=0;
    
    /* check phy status every 100 send calls */
    send_counter++;
    if (send_counter % 100) ifx_switch_check_phy_status();
 
    if( (incaip2_dma_read_channel_control_reg(1) & 0x00000001) != 0x00000001)
    {
        incaip2_dma_enable_channels();
    }
    if (length <= 0) 
    {
        printf("%s: bad packet size: %d\n", dev->name, length);
        goto Done;
    }
    
    
    if(tx_desc->status.field.OWN==DMA_OWN)
    {
        printf("no TX desc for CPU, drop packet\n");
        return 0;
    }
    
    /*
     * wait for descriptor completion
     */
    for(i = 0; tx_desc->status.field.C == 0; i++) 
    {
        if (i >= TOUT_LOOP) 
        {
            printf("%s: tx error buffer not ready\n", dev->name);
            goto Done;
        }
    }
     
    tx_desc->Data_Pointer=(u32)PHYSADDR((u32)packet)-byte_offset;
    
    tx_desc->status.word = DMA_DESC_OWN_DMA | 
                             DMA_DESC_SOP_SET |
                             DMA_DESC_EOP_SET |
                             byte_offset<<DMA_DESC_BYTEOFF_SHIFT |
                             length;    
    *((u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)=0x00000001;
    tx_chan_interrupts = *((u32*)CHANNEL_RELATED_REGISTERS_DMA_CIS);
    if(tx_chan_interrupts & DESCR_UNDERRUN)
    {
        *((u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)=0x00000001;
        *((u32*)CHANNEL_RELATED_REGISTERS_DMA_CIS) = DESCR_UNDERRUN;
    }
    /* 
     * increment tx descriptor 
     */
    curr_tx_descr = (curr_tx_descr+1) % NUM_TX_DESC;
    res = length;
Done:
    return res;
}


static int incaip2_switch_recv(struct eth_device *dev)
{
    int            length  = 0;
    rx_desc_t*     rx_desc;
    u32 rx_chan_interrupts;    

    for (;;) 
    {
        /*
         * get current rx descriptor and check control bits for data reception 
         */
        rx_desc = KSEG1ADDR(&rx_ring[curr_rx_descr]);
        
        if (rx_desc->status.field.C == 0) 
        {
            /* nothing received -> leave the function */
            break;
        }
        if (! rx_desc->status.field.EoP) 
        {
            /* packet only partly received -> try again later */
            printf("Partly received packet!!!\n");
            break;
        }

        /* get number of received bytes */
        length = rx_desc->status.field.data_length;

        if (length) 
        {
            NetReceive((void*)KSEG1ADDR(NetRxPackets[curr_rx_descr]), length - 4);
        }
        else 
        {
            printf("Zero length!!!\n");
        }
        /* free descriptor for DMA usage and proceed with next descriptor */
        rx_desc->status.field.EoP = 0;
        rx_desc->status.field.SoP = 0;        
        rx_desc->status.field.C   = 0; 
        rx_desc->status.field.data_length=PKTSIZE_ALIGN;        
        rx_desc->status.field.OWN = DMA_OWN;        
        curr_rx_descr = (curr_rx_descr + 1) % NUM_RX_DESC;
 
        *((u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)=0x00000000;
        rx_chan_interrupts = *((u32*)CHANNEL_RELATED_REGISTERS_DMA_CIS);
        if(rx_chan_interrupts & DESCR_UNDERRUN)
        {
            *((u32*)CHANNEL_RELATED_REGISTERS_DMA_CS)=0x00000000;
            *((u32*)CHANNEL_RELATED_REGISTERS_DMA_CIS) = DESCR_UNDERRUN;
        }
    }

    return length;

}


static void incaip2_switch_halt(struct eth_device *dev)
{
    u8 own_mac_addr[6];
    BOOL lan_port; 
    BOOL pc_port;
    BOOL cpu_port;    
    BOOL criticalFlag;  
    u32 maState;

    /* do not forward unknowns/broadcasts to CPU port */
    INCAIP2_SW_REG32( UNKNOWN_DEST_REG ) = 0x000000DB;

    /* Invalidate own MAC address which is passed in eth device structure */
    memcpy(own_mac_addr, dev->enetaddr, 6);
    if( incaip2_mac_table_add_entry( own_mac_addr,
                                     lan_port = FALSE, 
                                     pc_port = FALSE, 
                                     cpu_port = FALSE,
                                     criticalFlag = FALSE,
                                     maState = INVALID ) != 0)
    {
        printf("%s - Invalidating MAC table entry failed\n", __FUNCTION__);
    }
    /* Disable Switch CPU Port, DMA channels & interrupts */
    INCAIP2_SW_REG32( ST_PT_REG ) = 0x0000000F;
    incaip2_dma_disable_channel_interrupts();
    incaip2_dma_disable_channels();
}


static void incaip2_switch_init_chip(void)
{
    u16 temp_reg;

    /* Drop on CPU port, Learn-and-Forward on ports 0 and 1 */
    INCAIP2_SW_REG32( ST_PT_REG ) = 0x0000000F;

    /* forward broadcast and unknown multicast/unicast only to PC/LAN ports */
    INCAIP2_SW_REG32( UNKNOWN_DEST_REG ) = 0x000000DB;

    /* add CRC for packets from DMA to PMAC. Add TAG is disabled. */
    INCAIP2_SW_REG32( SWITCH_PMAC_HD_CTL ) = 0x00040000;
   
#if defined(CONFIG_PHY_MV88E1111) || defined(CONFIG_PHY_MV88E1121)
    /*
     * use external PHYs, disable MDIO hardware state machine for the 
     * following PHY register configurations.
     */   
    INCAIP2_SW_REG32( MDIO_MOD_SEL ) = 0x00000020;

    /* Set Switch Gigabit Mode Selection bit for LAN and PC port */
    INCAIP2_SW_REG32( SWITCH_P0_GMAC_REG ) = 0x0000A009;
    INCAIP2_SW_REG32( SWITCH_P1_GMAC_REG ) = 0x0000A009;

#if defined(CONFIG_PHY_MV88E1121)
    /* Port0=LAN gets ID1, Port1=PC gets ID0 (flaw on reference board) */
    INCAIP2_SW_REG32( SWITCH_MDIO_ID ) = 0x00000001;

/* Add delay to Rx and Tx Clock w.r.t. data */
    INCAIP2_MDIO_WRITE(PHY_LAN, 22, 2 );       /* select page 2 */
    INCAIP2_MDIO_READ(PHY_LAN, 21, temp_reg);
    INCAIP2_MDIO_WRITE(PHY_LAN, 21, temp_reg | 0x30);
    INCAIP2_MDIO_WRITE(PHY_LAN, 22, 0 );       /* select page 0 */
    INCAIP2_MDIO_WRITE(PHY_PC, 22, 2 );       /* select page 2 */
    INCAIP2_MDIO_READ(PHY_PC, 21, temp_reg);
    INCAIP2_MDIO_WRITE(PHY_PC, 21, temp_reg | 0x30);
    INCAIP2_MDIO_WRITE(PHY_PC, 22, 0 );       /* select page 0 */
#elif defined ( CONFIG_PHY_MV88E1111 ) 
    /* Add delay to Rx and Tx Clock w.r.t. data */
    INCAIP2_MDIO_READ(PHY_LAN, 20, temp_reg);
    INCAIP2_MDIO_WRITE(PHY_LAN, 20, temp_reg | 0x82);
    INCAIP2_MDIO_READ(PHY_PC, 20, temp_reg);
    INCAIP2_MDIO_WRITE(PHY_PC, 20, temp_reg | 0x82);
#endif

    /* Configure PHY Control Register for Port 0
            (B100)                 -- PHY control register --
       (8000:8000) 0.15    =     1    reset
       (4000:0000) 0.14    =     0    loopback
       (2040:2000) 0. 6,13 =   b01    speed selection = 100 Mbps
       (1000:1000) 0.12    =     1    A/N enable
       (0800:0000) 0.11    =     0    power-down
       (0400:0000) 0.10    =     0    isolate
       (0200:0000) 0. 9    =     0    restart A/N
       (0100:0100) 0. 8    =     1    duplex = full
       (0080:0000) 0. 7    =     0    collision test enable
       (003f:0000) 0. 5- 0 =     0    (reserved)
     */
    INCAIP2_MDIO_WRITE(PHY_LAN, 0, 0x3100);
    /* Soft-reset PHY0 */
    INCAIP2_MDIO_WRITE(PHY_LAN, 0, 0xb100);

    /* ... same configuration for Port 1 */
    INCAIP2_MDIO_WRITE(PHY_PC, 0, 0x3100);
    /* Soft-reset PHY1 */
    INCAIP2_MDIO_WRITE(PHY_PC, 0, 0xb100);

    /* dummy read needed because MV88E1111/MV88E1121 latches the Link Status bit */
    INCAIP2_MDIO_READ( PHY_LAN, 1, temp_reg);
    INCAIP2_MDIO_READ( PHY_PC, 1, temp_reg);

#else /* use INCA-IP2 internal PHYs */

    /*
     * use internal PHYs, disable MDIO hardware state machine for the 
     * following PHY register configurations.
     */   
    INCAIP2_SW_REG32( MDIO_MOD_SEL ) = 0x00000000;

    /* Configure LAN and PC port MACs to operate in 100Mbit MII mode  */
    INCAIP2_SW_REG32( SWITCH_P0_GMAC_REG ) = 0x0000A06A;
    INCAIP2_SW_REG32( SWITCH_P1_GMAC_REG ) = 0x0000A06A;

    /*
     * Configure PHY Control Register for Port 0
     * - RST = 0 -> no reset
     * - LPBK = 0 -> no loopback mode
     * - SPD_L = 1 speed (low bit) = 100 Mbit/s
     * - ANEN = 1 -> autonegotiation enable
     * - PDN = 0 -> no power down, normal operation
     * - ISO = 0 -> no isolation from network
     * - RSTAR = 0 -> no restart of ANEG
     * - DPLX = 1 -> Full Duplex
     * - COLTST = 0 -> Collision Test disable
     */
    INCAIP2_MDIO_WRITE(PHY_LAN, 0, 0x3100);
    /* ... same configuration for Port 1 */
    INCAIP2_MDIO_WRITE(PHY_PC, 0, 0x3100);

#endif

    /* enable ANEG and MDIO HW state machine */
    INCAIP2_SW_REG32(MDIO_MOD_SEL) |= 0x00000007;

    /* forward unknown broadcast/multicast/unicast to all ports */
    INCAIP2_SW_REG32(UNKNOWN_DEST_REG)=0x000001FF;

    /* Configure PMAC to add CRC in packet */
    INCAIP2_SW_REG32(SWITCH_PMAC_HD_CTL) = 0x00040000;  

    /* workaround for aging bug 
     * Since the hardware driven mac table aging is ~15 times faster than desired
     * it is configured to 1 day aging period which results in appr. 2:30 hours
     */
    INCAIP2_SW_REG32(PR_CTRL_REG) = (INCAIP2_SW_REG32(PR_CTRL_REG) & 0xFFFFC7FF) | 0x00003000;
    
    /* additionally enable CPU port in forwarding mode */    
    INCAIP2_SW_REG32(ST_PT_REG)=0x0000003F;

}

int incaip2_ethernet_initialize(bd_t * bis)
{
    struct eth_device *dev;

    if (!(dev = (struct eth_device *) malloc (sizeof *dev))) 
    {
        printf("Failed to allocate memory\n");
        return 0;
    }
    memset(dev, 0, sizeof(*dev));
    
    incaip2_dma_init();
    incaip2_switch_init_chip();    

#if defined ( CONFIG_PHY_MV88E1111 )
    sprintf(dev->name, "incaip2/88e1111");
#elif defined ( CONFIG_PHY_MV88E1121 )
    sprintf(dev->name, "incaip2/88e1121");
#else
    sprintf(dev->name, "incaip2");
#endif
    
    dev->init = incaip2_switch_init;
    dev->halt = incaip2_switch_halt;
    dev->send = incaip2_switch_send;
    dev->recv = incaip2_switch_recv;

    eth_register(dev);
    /* 
     * To workaround fast MAC address aging bug set table aging value to maximum.
     * Maximum is 1 day aging time which is ~150 minutes in reality.
     */
    INCAIP2_SW_REG32(PR_CTRL_REG) = (INCAIP2_SW_REG32(PR_CTRL_REG) & 0xFFFFC7FF) | 0x00003000;
    
    incaip2_dma_enable_channels();

    return 1;
}
#endif
